import numpy as np
from warnings import warn
from scipy.spatial.distance import cdist
from menpo.shape.base import Shape


class PointCloud(Shape):
    r"""
    An N-dimensional point cloud. This is internally represented as an `ndarray`
    of shape ``(n_points, n_dims)``. This class is important for dealing
    with complex functionality such as viewing and representing metadata such
    as landmarks.

    Currently only 2D and 3D pointclouds are viewable.

    Parameters
    ----------
    points : ``(n_points, n_dims)`` `ndarray`
        The array representing the points.
    copy : `bool`, optional
        If ``False``, the points will not be copied on assignment. Note that
        this will miss out on additional checks. Further note that we still
        demand that the array is C-contiguous - if it isn't, a copy will be
        generated anyway.
        In general this should only be used if you know what you are doing.
    """

    def __init__(self, points, copy=True):
        super(PointCloud, self).__init__()
        if not copy:
            if not points.flags.c_contiguous:
                warn('The copy flag was NOT honoured. A copy HAS been made. '
                     'Please ensure the data you pass is C-contiguous.')
                points = np.array(points, copy=True, order='C')
        else:
            points = np.array(points, copy=True, order='C')
        self.points = points

    @property
    def n_points(self):
        r"""
        The number of points in the pointcloud.

        :type: `int`
        """
        return self.points.shape[0]

    @property
    def n_dims(self):
        r"""
        The number of dimensions in the pointcloud.

        :type: `int`
        """
        return self.points.shape[1]

    def h_points(self):
        r"""
        Convert poincloud to a homogeneous array: ``(n_dims + 1, n_points)``

        :type: ``type(self)``
        """
        return np.concatenate((self.points.T, np.ones(self.n_points)[None, :]))

    def centre(self):
        r"""
        The mean of all the points in this PointCloud (centre of mass).

        Returns
        -------
        centre : ``(n_dims)`` `ndarray`
            The mean of this PointCloud's points.
        """
        return np.mean(self.points, axis=0)

    def centre_of_bounds(self):
        r"""
        The centre of the absolute bounds of this PointCloud. Contrast with
        :meth:`centre`, which is the mean point position.

        Returns
        -------
        centre : ``n_dims`` `ndarray`
            The centre of the bounds of this PointCloud.
        """
        min_b, max_b = self.bounds()
        return (min_b + max_b) / 2

    def _as_vector(self):
        r"""
        Returns a flattened representation of the pointcloud.
        Note that the flattened representation is of the form
        ``[x0, y0, x1, y1, ....., xn, yn]`` for 2D.

        Returns
        -------
        flattened : ``(n_points,)`` `ndarray`
            The flattened points.
        """
        return self.points.ravel()

    def tojson(self):
        r"""
        Convert this :map:`PointCloud` to a dictionary representation suitable
        for inclusion in the LJSON landmark format.

        Returns
        -------
        json : `dict`
            Dictionary with ``points`` keys.
        """
        return {'points': self.points.tolist()}

    def from_vector_inplace(self, vector):
        r"""
        Updates the points of this PointCloud in-place with the reshaped points
        from the provided vector. Note that the vector should have the form
        ``[x0, y0, x1, y1, ....., xn, yn]`` for 2D.

        Parameters
        ----------
        vector : ``(n_points,)`` `ndarray`
            The vector from which to create the points' array.
        """
        self.points = vector.reshape([-1, self.n_dims])

    def __str__(self):
        return '{}: n_points: {}, n_dims: {}'.format(type(self).__name__,
                                                     self.n_points,
                                                     self.n_dims)

    def bounds(self, boundary=0):
        r"""
        The minimum to maximum extent of the PointCloud. An optional boundary
        argument can be provided to expand the bounds by a constant margin.

        Parameters
        ----------
        boundary : `float`
            A optional padding distance that is added to the bounds. Default
            is ``0``, meaning the max/min of tightest possible containing
            square/cube/hypercube is returned.

        Returns
        -------
        min_b : ``(n_dims,)`` `ndarray`
            The minimum extent of the :map:`PointCloud` and boundary along
            each dimension
        max_b : ``(n_dims,)`` `ndarray`
            The maximum extent of the :map:`PointCloud` and boundary along
            each dimension
        """
        min_b = np.min(self.points, axis=0) - boundary
        max_b = np.max(self.points, axis=0) + boundary
        return min_b, max_b

    def range(self, boundary=0):
        r"""
        The range of the extent of the PointCloud.

        Parameters
        ----------
        boundary : `float`
            A optional padding distance that is used to extend the bounds
            from which the range is computed. Default is ``0``, no extension
            is performed.

        Returns
        -------
        range : ``(n_dims,)`` `ndarray`
            The range of the :map:`PointCloud` extent in each dimension.
        """
        min_b, max_b = self.bounds(boundary)
        return max_b - min_b

    def bounding_box(self):
        r"""
        Return the bounding box of this PointCloud as a directed graph.
        The the first point (0) will be nearest the origin for an axis aligned
        Pointcloud.
        In the case of an image, this ordering would appear as:

        ::

            0<--3
            |   ^
            |   |
            v   |
            1-->2

        Returns
        -------
        bounding_box : :map:`PointDirectedGraph`
            The axis aligned bounding box of the PointCloud.
        """
        from .graph import PointDirectedGraph
        min_p, max_p = self.bounds()
        return PointDirectedGraph(np.array([min_p, [max_p[0], min_p[1]],
                                            max_p, [min_p[0], max_p[1]]]),
                                  np.array([[0, 1], [1, 2], [2, 3], [3, 0]]),
                                  copy=False)

    def _view_2d(self, figure_id=None, new_figure=False, image_view=True,
                 render_markers=True, marker_style='o', marker_size=20,
                 marker_face_colour='r', marker_edge_colour='k',
                 marker_edge_width=1., render_axes=True,
                 axes_font_name='sans-serif', axes_font_size=10,
                 axes_font_style='normal', axes_font_weight='normal',
                 axes_x_limits=None, axes_y_limits=None, figure_size=(10, 8),
                 label=None, **kwargs):
        r"""
        Visualization of the PointCloud in 2D.

        Returns
        -------
        figure_id : `object`, optional
            The id of the figure to be used.
        new_figure : `bool`, optional
            If ``True``, a new figure is created.
        image_view : `bool`, optional
            If ``True`` the PointCloud will be viewed as if it is in the image
            coordinate system.
        render_lines : `bool`, optional
            If ``True``, the edges will be rendered.
        line_colour : See Below, optional
            The colour of the lines.
            Example options::

                {r, g, b, c, m, k, w}
                or
                (3, ) ndarray

        line_style : ``{-, --, -., :}``, optional
            The style of the lines.
        line_width : `float`, optional
            The width of the lines.
        render_markers : `bool`, optional
            If ``True``, the markers will be rendered.
        marker_style : See Below, optional
            The style of the markers. Example options ::

                {., ,, o, v, ^, <, >, +, x, D, d, s, p, *, h, H, 1, 2, 3, 4, 8}

        marker_size : `int`, optional
            The size of the markers in points^2.
        marker_face_colour : See Below, optional
            The face (filling) colour of the markers.
            Example options ::

                {r, g, b, c, m, k, w}
                or
                (3, ) ndarray

        marker_edge_colour : See Below, optional
            The edge colour of the markers.
            Example options ::

                {r, g, b, c, m, k, w}
                or
                (3, ) ndarray

        marker_edge_width : `float`, optional
            The width of the markers' edge.
        render_axes : `bool`, optional
            If ``True``, the axes will be rendered.
        axes_font_name : See Below, optional
            The font of the axes.
            Example options ::

                {serif, sans-serif, cursive, fantasy, monospace}

        axes_font_size : `int`, optional
            The font size of the axes.
        axes_font_style : {``normal``, ``italic``, ``oblique``}, optional
            The font style of the axes.
        axes_font_weight : See Below, optional
            The font weight of the axes.
            Example options ::

                {ultralight, light, normal, regular, book, medium, roman,
                semibold, demibold, demi, bold, heavy, extra bold, black}

        axes_x_limits : (`float`, `float`) `tuple` or ``None``, optional
            The limits of the x axis.
        axes_y_limits : (`float`, `float`) `tuple` or ``None``, optional
            The limits of the y axis.
        figure_size : (`float`, `float`) `tuple` or ``None``, optional
            The size of the figure in inches.
        label : `str`, optional
            The name entry in case of a legend.

        Returns
        -------
        viewer : :map:`PointGraphViewer2d`
            The viewer object.
        """
        from menpo.visualize.base import PointGraphViewer2d
        adjacency_array = np.empty(0)
        renderer = PointGraphViewer2d(figure_id, new_figure,
                                      self.points,
                                      adjacency_array)
        renderer.render(
            image_view=image_view, render_lines=False, line_colour='b',
            line_style='-', line_width=1., render_markers=render_markers,
            marker_style=marker_style, marker_size=marker_size,
            marker_face_colour=marker_face_colour,
            marker_edge_colour=marker_edge_colour,
            marker_edge_width=marker_edge_width, render_axes=render_axes,
            axes_font_name=axes_font_name, axes_font_size=axes_font_size,
            axes_font_style=axes_font_style, axes_font_weight=axes_font_weight,
            axes_x_limits=axes_x_limits, axes_y_limits=axes_y_limits,
            figure_size=figure_size, label=label)
        return renderer

    def _view_landmarks_2d(self, group=None, with_labels=None,
                           without_labels=None, figure_id=None,
                           new_figure=False, image_view=True, render_lines=True,
                           line_colour=None, line_style='-', line_width=1,
                           render_markers=True, marker_style='o',
                           marker_size=20, marker_face_colour=None,
                           marker_edge_colour=None, marker_edge_width=1.,
                           render_numbering=False,
                           numbers_horizontal_align='center',
                           numbers_vertical_align='bottom',
                           numbers_font_name='sans-serif', numbers_font_size=10,
                           numbers_font_style='normal',
                           numbers_font_weight='normal',
                           numbers_font_colour='k', render_legend=False,
                           legend_title='', legend_font_name='sans-serif',
                           legend_font_style='normal', legend_font_size=10,
                           legend_font_weight='normal',
                           legend_marker_scale=None, legend_location=2,
                           legend_bbox_to_anchor=(1.05, 1.),
                           legend_border_axes_pad=None, legend_n_columns=1,
                           legend_horizontal_spacing=None,
                           legend_vertical_spacing=None, legend_border=True,
                           legend_border_padding=None, legend_shadow=False,
                           legend_rounded_corners=False, render_axes=False,
                           axes_font_name='sans-serif', axes_font_size=10,
                           axes_font_style='normal', axes_font_weight='normal',
                           axes_x_limits=None, axes_y_limits=None,
                           figure_size=(10, 8)):
        """
        Visualize the landmarks. This method will appear on the Image as
        ``view_landmarks`` if the Image is 2D.

        Parameters
        ----------
        group : `str` or``None`` optional
            The landmark group to be visualized. If ``None`` and there are more
            than one landmark groups, an error is raised.
        with_labels : ``None`` or `str` or `list` of `str`, optional
            If not ``None``, only show the given label(s). Should **not** be
            used with the ``without_labels`` kwarg.
        without_labels : ``None`` or `str` or `list` of `str`, optional
            If not ``None``, show all except the given label(s). Should **not**
            be used with the ``with_labels`` kwarg.
        figure_id : `object`, optional
            The id of the figure to be used.
        new_figure : `bool`, optional
            If ``True``, a new figure is created.
        image_view : `bool`, optional
            If ``True`` the PointCloud will be viewed as if it is in the image
            coordinate system.
        render_lines : `bool`, optional
            If ``True``, the edges will be rendered.
        line_colour : See Below, optional
            The colour of the lines.
            Example options::

                {r, g, b, c, m, k, w}
                or
                (3, ) ndarray

        line_style : ``{-, --, -., :}``, optional
            The style of the lines.
        line_width : `float`, optional
            The width of the lines.
        render_markers : `bool`, optional
            If ``True``, the markers will be rendered.
        marker_style : See Below, optional
            The style of the markers. Example options ::

                {., ,, o, v, ^, <, >, +, x, D, d, s, p, *, h, H, 1, 2, 3, 4, 8}

        marker_size : `int`, optional
            The size of the markers in points^2.
        marker_face_colour : See Below, optional
            The face (filling) colour of the markers.
            Example options ::

                {r, g, b, c, m, k, w}
                or
                (3, ) ndarray

        marker_edge_colour : See Below, optional
            The edge colour of the markers.
            Example options ::

                {r, g, b, c, m, k, w}
                or
                (3, ) ndarray

        marker_edge_width : `float`, optional
            The width of the markers' edge.
        render_numbering : `bool`, optional
            If ``True``, the landmarks will be numbered.
        numbers_horizontal_align : ``{center, right, left}``, optional
            The horizontal alignment of the numbers' texts.
        numbers_vertical_align : ``{center, top, bottom, baseline}``, optional
            The vertical alignment of the numbers' texts.
        numbers_font_name : See Below, optional
            The font of the numbers. Example options ::

                {serif, sans-serif, cursive, fantasy, monospace}

        numbers_font_size : `int`, optional
            The font size of the numbers.
        numbers_font_style : ``{normal, italic, oblique}``, optional
            The font style of the numbers.
        numbers_font_weight : See Below, optional
            The font weight of the numbers.
            Example options ::

                {ultralight, light, normal, regular, book, medium, roman,
                semibold, demibold, demi, bold, heavy, extra bold, black}

        numbers_font_colour : See Below, optional
            The font colour of the numbers.
            Example options ::

                {r, g, b, c, m, k, w}
                or
                (3, ) ndarray

        render_legend : `bool`, optional
            If ``True``, the legend will be rendered.
        legend_title : `str`, optional
            The title of the legend.
        legend_font_name : See below, optional
            The font of the legend. Example options ::

                {serif, sans-serif, cursive, fantasy, monospace}

        legend_font_style : ``{normal, italic, oblique}``, optional
            The font style of the legend.
        legend_font_size : `int`, optional
            The font size of the legend.
        legend_font_weight : See Below, optional
            The font weight of the legend.
            Example options ::

                {ultralight, light, normal, regular, book, medium, roman,
                semibold, demibold, demi, bold, heavy, extra bold, black}

        legend_marker_scale : `float`, optional
            The relative size of the legend markers with respect to the original
        legend_location : `int`, optional
            The location of the legend. The predefined values are:

            =============== ==
            'best'          0
            'upper right'   1
            'upper left'    2
            'lower left'    3
            'lower right'   4
            'right'         5
            'center left'   6
            'center right'  7
            'lower center'  8
            'upper center'  9
            'center'        10
            =============== ==

        legend_bbox_to_anchor : (`float`, `float`) `tuple`, optional
            The bbox that the legend will be anchored.
        legend_border_axes_pad : `float`, optional
            The pad between the axes and legend border.
        legend_n_columns : `int`, optional
            The number of the legend's columns.
        legend_horizontal_spacing : `float`, optional
            The spacing between the columns.
        legend_vertical_spacing : `float`, optional
            The vertical space between the legend entries.
        legend_border : `bool`, optional
            If ``True``, a frame will be drawn around the legend.
        legend_border_padding : `float`, optional
            The fractional whitespace inside the legend border.
        legend_shadow : `bool`, optional
            If ``True``, a shadow will be drawn behind legend.
        legend_rounded_corners : `bool`, optional
            If ``True``, the frame's corners will be rounded (fancybox).
        render_axes : `bool`, optional
            If ``True``, the axes will be rendered.
        axes_font_name : See Below, optional
            The font of the axes. Example options ::

                {serif, sans-serif, cursive, fantasy, monospace}

        axes_font_size : `int`, optional
            The font size of the axes.
        axes_font_style : ``{normal, italic, oblique}``, optional
            The font style of the axes.
        axes_font_weight : See Below, optional
            The font weight of the axes.
            Example options ::

                {ultralight, light, normal, regular, book, medium, roman,
                semibold,demibold, demi, bold, heavy, extra bold, black}

        axes_x_limits : (`float`, `float`) `tuple` or ``None`` optional
            The limits of the x axis.
        axes_y_limits : (`float`, `float`) `tuple` or ``None`` optional
            The limits of the y axis.
        figure_size : (`float`, `float`) `tuple` or ``None`` optional
            The size of the figure in inches.

        Raises
        ------
        ValueError
            If both ``with_labels`` and ``without_labels`` are passed.
        ValueError
            If the landmark manager doesn't contain the provided group label.
        """
        if not self.has_landmarks:
            raise ValueError('PointCloud does not have landmarks attached, '
                             'unable to view landmarks.')
        self_view = self.view(figure_id=figure_id, new_figure=new_figure,
                              image_view=image_view, figure_size=figure_size)
        landmark_view = self.landmarks[group].view(
            with_labels=with_labels, without_labels=without_labels,
            figure_id=self_view.figure_id, new_figure=False,
            image_view=image_view, render_lines=render_lines,
            line_colour=line_colour, line_style=line_style,
            line_width=line_width, render_markers=render_markers,
            marker_style=marker_style, marker_size=marker_size,
            marker_face_colour=marker_face_colour,
            marker_edge_colour=marker_edge_colour,
            marker_edge_width=marker_edge_width,
            render_numbering=render_numbering,
            numbers_horizontal_align=numbers_horizontal_align,
            numbers_vertical_align=numbers_vertical_align,
            numbers_font_name=numbers_font_name,
            numbers_font_size=numbers_font_size,
            numbers_font_style=numbers_font_style,
            numbers_font_weight=numbers_font_weight,
            numbers_font_colour=numbers_font_colour,
            render_legend=render_legend, legend_title=legend_title,
            legend_font_name=legend_font_name,
            legend_font_style=legend_font_style,
            legend_font_size=legend_font_size,
            legend_font_weight=legend_font_weight,
            legend_marker_scale=legend_marker_scale,
            legend_location=legend_location,
            legend_bbox_to_anchor=legend_bbox_to_anchor,
            legend_border_axes_pad=legend_border_axes_pad,
            legend_n_columns=legend_n_columns,
            legend_horizontal_spacing=legend_horizontal_spacing,
            legend_vertical_spacing=legend_vertical_spacing,
            legend_border=legend_border,
            legend_border_padding=legend_border_padding,
            legend_shadow=legend_shadow,
            legend_rounded_corners=legend_rounded_corners,
            render_axes=render_axes, axes_font_name=axes_font_name,
            axes_font_size=axes_font_size, axes_font_style=axes_font_style,
            axes_font_weight=axes_font_weight, axes_x_limits=axes_x_limits,
            axes_y_limits=axes_y_limits, figure_size=figure_size)

        return landmark_view

    def _view_3d(self, figure_id=None, new_figure=False):
        r"""
        Visualization of the PointCloud in 3D.

        Parameters
        ----------
        figure_id : `object`, optional
            The id of the figure to be used.
        new_figure : `bool`, optional
            If ``True``, a new figure is created.

        Returns
        -------
        viewer : PointCloudViewer3d
            The Menpo3D viewer object.
        """
        try:
            from menpo3d.visualize import PointCloudViewer3d
            return PointCloudViewer3d(figure_id, new_figure,
                                      self.points).render()
        except ImportError:
            from menpo.visualize import Menpo3dErrorMessage
            raise ImportError(Menpo3dErrorMessage)

    def _view_landmarks_3d(self, figure_id=None, new_figure=False,
                           group=None):
        r"""
        Visualization of the PointCloud landmarks in 3D.

        Parameters
        ----------
        figure_id : `object`, optional
            The id of the figure to be used.
        new_figure : `bool`, optional
            If ``True``, a new figure is created.
        group : `str`
            The landmark group to visualize. If ``None`` is passed, the first
            and only landmark group on the object will be visualized.

        Returns
        -------
        viewer : LandmarkViewer3d
            The Menpo3D viewer object.
        """
        try:
            from menpo3d.visualize import LandmarkViewer3d
            self_renderer = self.view(figure_id=figure_id,
                                      new_figure=new_figure)
            return LandmarkViewer3d(self_renderer.figure, False,  self,
                                    self.landmarks[group]).render()
        except ImportError:
            from menpo.visualize import Menpo3dErrorMessage
            raise ImportError(Menpo3dErrorMessage)

    def view_widget(self, popup=False, browser_style='buttons',
                    figure_size=(10, 8)):
        r"""
        Visualization of the PointCloud using the :map:`visualize_pointclouds`
        widget.

        Parameters
        ----------
        popup : `bool`, optional
            If ``True``, the widget will be rendered in a popup window.
        browser_style : ``{buttons, slider}``, optional
            It defines whether the selector of the PointCloud objects will have
            the form of plus/minus buttons or a slider.
        figure_size : (`int`, `int`), optional
            The initial size of the rendered figure.
        """
        from menpo.visualize import visualize_pointclouds
        visualize_pointclouds(self, popup=popup, figure_size=figure_size,
                              browser_style=browser_style)

    def _transform_self_inplace(self, transform):
        self.points = transform(self.points)
        return self

    def distance_to(self, pointcloud, **kwargs):
        r"""
        Returns a distance matrix between this PointCloud and another.
        By default the Euclidean distance is calculated - see
        `scipy.spatial.distance.cdist` for valid kwargs to change the metric
        and other properties.

        Parameters
        ----------
        pointcloud : :map:`PointCloud`
            The second pointcloud to compute distances between. This must be
            of the same dimension as this PointCloud.

        Returns
        -------
        distance_matrix: ``(n_points, n_points)`` `ndarray`
            The symmetric pairwise distance matrix between the two PointClouds
            s.t. ``distance_matrix[i, j]`` is the distance between the i'th
            point of this PointCloud and the j'th point of the input
            PointCloud.
        """
        if self.n_dims != pointcloud.n_dims:
            raise ValueError("The two PointClouds must be of the same "
                             "dimensionality.")
        return cdist(self.points, pointcloud.points, **kwargs)

    def norm(self, **kwargs):
        r"""
        Returns the norm of this PointCloud. This is a translation and
        rotation invariant measure of the point cloud's intrinsic size - in
        other words, it is always taken around the point cloud's centre.

        By default, the Frobenius norm is taken, but this can be changed by
        setting kwargs - see ``numpy.linalg.norm`` for valid options.

        Returns
        -------
        norm : `float`
            The norm of this :map:`PointCloud`
        """
        return np.linalg.norm(self.points - self.centre(), **kwargs)

    def from_mask(self, mask):
        """
        A 1D boolean array with the same number of elements as the number of
        points in the PointCloud. This is then broadcast across the dimensions
        of the PointCloud and returns a new PointCloud containing only those
        points that were ``True`` in the mask.

        Parameters
        ----------
        mask : ``(n_points,)`` `ndarray`
            1D array of booleans

        Returns
        -------
        pointcloud : :map:`PointCloud`
            A new pointcloud that has been masked.

        Raises
        ------
        ValueError
            Mask must have same number of points as pointcloud.
        """
        if mask.shape[0] != self.n_points:
            raise ValueError('Mask must be a 1D boolean array of the same '
                             'number of entries as points in this PointCloud.')
        pc = self.copy()
        pc.points = pc.points[mask, :]
        return pc



#cython: cdivision=True
#cython: boundscheck=False
#cython: nonecheck=False
#cython: wraparound=False
import numpy as np

cimport numpy as cnp
from .interpolation cimport (nearest_neighbour_interpolation,
                             bilinear_interpolation,
                             biquadratic_interpolation,
                             bicubic_interpolation)


cdef inline void _matrix_transform(double x, double y, double* H, double *x_,
                                   double *y_):
    """Apply a homography to a coordinate.

    Parameters
    ----------
    x, y : double
        Input coordinate.
    H : (3,3) *double
        Transformation matrix.
    x_, y_ : *double
        Output coordinate.

    """
    cdef double xx, yy, zz

    xx = H[0] * x + H[1] * y + H[2]
    yy = H[3] * x + H[4] * y + H[5]
    zz =  H[6] * x + H[7] * y + H[8]

    x_[0] = xx / zz
    y_[0] = yy / zz


def _warp_fast(cnp.ndarray image, cnp.ndarray H, output_shape=None,
               int order=1, mode='constant', double cval=0):
    """Projective transformation (homography).

    Perform a projective transformation (homography) of a
    floating point image, using interpolation.

    For each pixel, given its homogeneous coordinate :math:`\mathbf{x}
    = [x, y, 1]^T`, its target position is calculated by multiplying
    with the given matrix, :math:`H`, to give :math:`H \mathbf{x}`.
    E.g., to rotate by theta degrees clockwise, the matrix should be::

      [[cos(theta) -sin(theta) 0]
       [sin(theta)  cos(theta) 0]
       [0            0         1]]

    or, to translate x by 10 and y by 20::

      [[1 0 10]
       [0 1 20]
       [0 0 1 ]].

    Parameters
    ----------
    image : 2-D array
        Input image.
    H : array of shape ``(3, 3)``
        Transformation matrix H that defines the homography.
    output_shape : tuple (rows, cols), optional
        Shape of the output image generated (default None).
    order : {0, 1, 2, 3}, optional
        Order of interpolation::
        * 0: Nearest-neighbor
        * 1: Bi-linear (default)
        * 2: Bi-quadratic
        * 3: Bi-cubic
    mode : {'constant', 'reflect', 'wrap', 'nearest'}, optional
        How to handle values outside the image borders (default is constant).
    cval : string, optional (default 0)
        Used in conjunction with mode 'C' (constant), the value
        outside the image boundaries.

    """

    cdef double[:, ::1] img = np.ascontiguousarray(image, dtype=np.double)
    cdef double[:, ::1] M = np.ascontiguousarray(H)

    if mode not in ('constant', 'wrap', 'reflect', 'nearest'):
        raise ValueError("Invalid mode specified.  Please use "
                         "`constant`, `nearest`, `wrap` or `reflect`.")
    cdef char mode_c = ord(mode[0].upper())

    cdef Py_ssize_t out_r, out_c
    if output_shape is None:
        out_r = int(img.shape[0])
        out_c = int(img.shape[1])
    else:
        out_r = int(output_shape[0])
        out_c = int(output_shape[1])

    cdef double[:, ::1] out = np.zeros((out_r, out_c), dtype=np.double)

    cdef Py_ssize_t tfr, tfc
    cdef double r, c
    cdef Py_ssize_t rows = img.shape[0]
    cdef Py_ssize_t cols = img.shape[1]

    cdef double (*interp_func)(double*, Py_ssize_t, Py_ssize_t, double, double,
                               char, double)
    if order == 0:
        interp_func = nearest_neighbour_interpolation
    elif order == 1:
        interp_func = bilinear_interpolation
    elif order == 2:
        interp_func = biquadratic_interpolation
    elif order == 3:
        interp_func = bicubic_interpolation

    for tfr in range(out_r):
        for tfc in range(out_c):
            _matrix_transform(tfc, tfr, &M[0, 0], &c, &r)
            out[tfr, tfc] = interp_func(&img[0, 0], rows, cols, r, c,
                                        mode_c, cval)

    return np.asarray(out)




import numpy as np
from numpy.testing import assert_allclose
from nose.tools import raises

from menpo.shape import UndirectedGraph, DirectedGraph, Tree

# Define adjacency arrays
adj_undirected = np.array([[0, 1],
                           [0, 2],
                           [1, 2],
                           [1, 3],
                           [2, 4],
                           [3, 4],
                           [3, 5]])
g_undirected = UndirectedGraph(adj_undirected)
adj_directed = np.array([[1, 0],
                         [2, 0],
                         [1, 2],
                         [2, 1],
                         [1, 3],
                         [2, 4],
                         [3, 4],
                         [3, 5]])
g_directed = DirectedGraph(adj_directed)
adj_tree = np.array([[0, 1],
                     [0, 2],
                     [1, 3],
                     [1, 4],
                     [2, 5],
                     [3, 6],
                     [4, 7],
                     [5, 8]])
g_tree = Tree(adj_tree, 0)


@raises(ValueError)
def test_create_graph_exception():
    adj = np.array([[0, 1, 3],
                    [0, 2, 4]])
    UndirectedGraph(adj)


@raises(ValueError)
def test_create_tree_exception_1():
    adj = np.array([[0, 1],
                    [0, 2],
                    [1, 3],
                    [1, 4],
                    [2, 5],
                    [3, 6],
                    [4, 7],
                    [5, 8],
                    [0, 4]])
    Tree(adj, 0)


@raises(ValueError)
def test_create_tree_exception_2():
    Tree(adj_tree, 20)


def test_n_edges():
    assert_allclose(g_directed.n_edges, 8)
    assert_allclose(g_undirected.n_edges, 7)
    assert_allclose(g_tree.n_edges, 8)


def test_n_vertices():
    assert_allclose(g_directed.n_vertices, 6)
    assert_allclose(g_undirected.n_vertices, 6)
    assert_allclose(g_tree.n_vertices, 9)


def test_adjacency_list():
    assert (g_directed.adjacency_list ==
            [[], [0, 2, 3], [0, 1, 4], [4, 5], [], []])
    assert (g_undirected.adjacency_list ==
            [[1, 2], [0, 2, 3], [0, 1, 4], [1, 4, 5], [2, 3], [3]])
    assert (g_tree.adjacency_list ==
            [[1, 2], [3, 4], [5], [6], [7], [8], [], [], []])


def test_adjacency_matrix():
    assert_allclose(g_directed.get_adjacency_matrix(),
                    np.array([[False, False, False, False, False, False],
                              [True, False, True, True, False, False],
                              [True, True, False, False, True, False],
                              [False, False, False, False, True, True],
                              [False, False, False, False, False, False],
                              [False, False, False, False, False, False]]))
    assert_allclose(g_undirected.get_adjacency_matrix(),
                    np.array([[False, True, True, False, False, False],
                              [True, False, True, True, False, False],
                              [True, True, False, False, True, False],
                              [False, True, False, False, True, True],
                              [False, False, True, True, False, False],
                              [False, False, False, True, False, False]]))
    assert_allclose(g_tree.get_adjacency_matrix(),
                    np.array([[False, True, True, False, False, False, False, False, False],
                              [False, False, False, True, True, False, False, False, False],
                              [False, False, False, False, False, True, False, False, False],
                              [False, False, False, False, False, False, True, False, False],
                              [False, False, False, False, False, False, False, True, False],
                              [False, False, False, False, False, False, False, False, True],
                              [False, False, False, False, False, False, False, False, False],
                              [False, False, False, False, False, False, False, False, False],
                              [False, False, False, False, False, False, False, False, False]]))


def test_n_paths():
    assert_allclose(g_directed.n_paths(1, 4), 2)
    assert_allclose(g_undirected.n_paths(0, 5), 4)
    assert_allclose(g_tree.n_paths(2, 6), 0)


def test_find_all_paths():
    assert (g_directed.find_all_paths(1, 4) == [[1, 2, 4], [1, 3, 4]])
    assert (g_undirected.find_all_paths(0, 5) ==
            [[0, 1, 2, 4, 3, 5], [0, 1, 3, 5], [0, 2, 1, 3, 5], [0, 2, 4, 3, 5]])
    assert (g_tree.find_all_paths(2, 6) == [])


def test_find_path():
    assert (g_directed.find_all_paths(1, 4)[0] == g_directed.find_path(1, 4))
    assert (g_undirected.find_all_paths(0, 5)[0] == g_undirected.find_path(0, 5))
    assert (g_tree.find_path(2, 6) is None)


def test_find_shortest_path():
    assert (g_directed.find_shortest_path(1, 0) == [1, 0])
    assert (g_undirected.find_shortest_path(5, 0) == [5, 3, 1, 0])
    assert (g_tree.find_shortest_path(1, 7) == [1, 4, 7])


def test_neighbours_children_parent():
    assert (g_directed.children(3) == [4, 5])
    assert (g_directed.n_children(0) == 0)
    assert (g_directed.parent(0) == [1, 2])
    assert (g_directed.n_parent(3) == 1)

    assert (g_undirected.neighbours(1) == [0, 2, 3])
    assert (g_undirected.n_neighbours(3) == 3)

    assert (g_tree.children(5) == [8])
    assert (g_tree.n_children(6) == 0)
    assert (g_tree.parent(4) == [1])


def test_tree():
    assert (g_tree.maximum_depth == 3)
    assert (g_tree.depth_of_vertex(0) == 0)
    assert (g_tree.depth_of_vertex(4) == 2)
    assert (g_tree.n_vertices_at_depth(2) == 3)
    assert g_tree.is_leaf(7)
    assert not g_tree.is_leaf(5)
    assert (g_tree.leaves == [6, 7, 8])
    assert (g_tree.n_leaves == 3)


def test_minimum_spanning_tree():
    adjacency_array = np.array([[3, 1],
                                [2, 3],
                                [0, 3],
                                [2, 0],
                                [0, 1]])
    weights = np.array([[0, 11, 13, 12],
                        [11, 0, 0, 14],
                        [13, 0, 0, 10],
                        [12, 14, 10, 0]])
    g = UndirectedGraph(adjacency_array)
    t = g.minimum_spanning_tree(weights, root_vertex=0)
    assert t.n_edges == 3
    assert_allclose(t.adjacency_array, np.array([[0, 1], [0, 3], [3, 2]]))
    assert t.adjacency_list == [[1, 3], [], [], [2]]
    assert_allclose(t.get_adjacency_matrix(),
                    np.array([[False, True, False, True],
                              [False, False, False, False],
                              [False, False, False, False],
                              [False, False, True, False]]))
    assert t.predecessors_list == [None, 0, 3, 0]


def test_is_edge():
    assert g_directed.is_edge(2, 4)
    assert not g_directed.is_edge(3, 1)
    assert not g_directed.is_edge(5, 0)

    assert g_undirected.is_edge(2, 1)
    assert g_undirected.is_edge(1, 2)
    assert not g_undirected.is_edge(5, 0)

    assert g_tree.is_edge(4, 7)
    assert not g_tree.is_edge(6, 3)
    assert not g_tree.is_edge(8, 1)


def test_is_tree():
    assert not g_undirected.is_tree()
    assert not g_directed.is_tree()
    assert g_tree.is_tree()



import cPickle as pickle
import gzip
from .base import Importer


class PickleImporter(Importer):

    def build(self):
        with open(self.filepath, 'rb') as f:
            x = pickle.load(f)
        return x


class GZipPickleImporter(Importer):

    def build(self):
        with gzip.open(self.filepath, 'rb') as f:
            x = pickle.load(f)
        return x



import sys


def progress_bar_str(percentage, bar_length=20, bar_marker='=', show_bar=True):
    r"""
    Returns an `str` of the specified progress percentage. The percentage is
    represented either in the form of a progress bar or in the form of a
    percentage number. It can be combined with the :func:`print_dynamic`
    function.

    Parameters
    ----------
    percentage : `float`
        The progress percentage to be printed. It must be in the range
        ``[0, 1]``.
    bar_length : `int`, optional
        Defines the length of the bar in characters.
    bar_marker : `str`, optional
        Defines the marker character that will be used to fill the bar.
    show_bar : `bool`, optional
        If ``True``, the `str` includes the bar followed by the percentage,
        e.g. ``'[=====     ] 50%'``

        If ``False``, the `str` includes only the percentage,
        e.g. ``'50%'``

    Returns
    -------
    progress_str : `str`
        The progress percentage string that can be printed.

    Raises
    ------
    ValueError
        ``percentage`` is not in the range ``[0, 1]``
    ValueError
        ``bar_length`` must be an integer >= ``1``
    ValueError
        ``bar_marker`` must be a string of length 1

    Examples
    --------
    This for loop: ::

        n_iters = 2000
        for k in range(n_iters):
            print_dynamic(progress_bar_str(float(k) / (n_iters-1)))

    prints a progress bar of the form: ::

        [=============       ] 68%
    """
    if percentage < 0:
        raise ValueError("percentage is not in the range [0, 1]")
    elif percentage > 1:
        percentage = 1
    if not isinstance(bar_length, int) or bar_length < 1:
        raise ValueError("bar_length must be an integer >= 1")
    if not isinstance(bar_marker, str) or len(bar_marker) != 1:
        raise ValueError("bar_marker must be a string of length 1")
    # generate output string
    if show_bar:
        str_param = "[%-" + str(bar_length) + "s] %d%%"
        bar_percentage = int(percentage * bar_length)
        return str_param % (bar_marker * bar_percentage, percentage * 100)
    else:
        return "%d%%" % (percentage * 100)


def print_dynamic(str_to_print):
    r"""
    Prints dynamically the provided `str`, i.e. the `str` is printed and then
    the buffer gets flushed.

    Parameters
    ----------
    str_to_print : `str`
        The string to print.
    """
    sys.stdout.write("\r%s" % str_to_print)
    sys.stdout.flush()


def print_bytes(num):
    r"""
    Converts bytes to a sensible format to be printed. For example: ::

        print_bytes(12345) returns '12.06 KB'
        print_bytes(123456789) returns '117.74 MB'

    Parameters
    ----------
    num : `int`
        The size in bytes.

    Raises
    ------
    ValueError
        num must be int >= 0
    """
    if not isinstance(num, int) or num < 0:
        raise ValueError("num must be int >= 0")
    for x in ['bytes', 'KB', 'MB', 'GB']:
        if num < 1024.0:
            return "{0:3.2f} {1:s}".format(num, x)
        num /= 1024.0
    return "{0:3.2f} {1:s}".format(num, 'TB')



import os


def _norm_path(filepath):
    r"""
    Uses all the tricks in the book to expand a path out to an absolute one.
    """
    return os.path.abspath(os.path.normpath(
        os.path.expandvars(os.path.expanduser(str(filepath)))))



from collections import OrderedDict
from functools import partial

import numpy as np

from .tools import (colour_selection, format_colour_selection,
                    hog_options, format_hog_options, igo_options,
                    format_igo_options, lbp_options,
                    format_lbp_options, daisy_options,
                    format_daisy_options, _convert_image_to_bytes,
                    line_options, format_line_options,
                    update_line_options, marker_options,
                    format_marker_options, update_marker_options,
                    numbering_options, format_numbering_options,
                    update_numbering_options, figure_options,
                    format_figure_options, update_figure_options,
                    figure_options_two_scales,
                    format_figure_options_two_scales,
                    update_figure_options_two_scales, grid_options,
                    format_grid_options, update_grid_options,
                    index_selection_slider, index_selection_buttons,
                    format_index_selection, update_index_selection,
                    legend_options, format_legend_options,
                    update_legend_options, image_options, format_image_options,
                    update_image_options)


def channel_options(channels_options_default, plot_function=None,
                    toggle_show_default=True, toggle_show_visible=True):
    r"""
    Creates a widget with Channel Options. Specifically, it has:
        1) Two radio buttons that select an options mode, depending on whether
           the user wants to visualize a "Single" or "Multiple" channels.
        2) If mode is "Single", the channel number is selected by one slider.
           If mode is "Multiple", the channel range is selected by two sliders.
        3) If mode is "Single" and the image has 3 channels, there is a checkbox
           option to visualize the image in RGB mode.
        4) If mode is "Multiple", there is a checkbox option to visualize the
           sum of the channels.
        5) If mode is "Multiple", there is a checkbox option to visualize the
           glyph.
        6) The glyph option is accompanied by a block size text field and a
           checkbox that enables negative values visualization.
        7) A checkbox that defines whether the masked image will be displayed.
        8) A toggle button that controls the visibility of all the above, i.e.
           the channel options.

    The structure of the widgets is the following:
        channel_options_wid.children = [toggle_button, all_but_toggle]
        all_but_toggle.children = [mode_and_masked, all_but_radio_buttons]
        mode_and_masked.children = [mode_radio_buttons, masked_checkbox]
        all_but_radio_buttons.children = [all_sliders, multiple_checkboxes]
        all_sliders.children = [first_slider, second_slider]
        multiple_checkboxes.children = [sum_checkbox, glyph_all, rgb_checkbox]
        glyph_all.children = [glyph_checkbox, glyph_options]
        glyph_options.children = [block_size_text, use_negative_checkbox]

    The returned widget saves the selected values in the following dictionary:
        channel_options_wid.selected_values

    To fix the alignment within this widget please refer to
    `format_channel_options()` function.

    To update the state of this widget, please refer to
    `update_channel_options()` function.

    Parameters
    ----------
    channels_options_default : `dict`
        The initial options. For example:
            channels_options_default = {'n_channels': 10,
                                        'image_is_masked': True,
                                        'channels': 0,
                                        'glyph_enabled': False,
                                        'glyph_block_size': 3,
                                        'glyph_use_negative': False,
                                        'sum_enabled': False,
                                        'masked_enabled': True}

    plot_function : `function` or None, optional
        The plot function that is executed when a widgets' value changes.
        If None, then nothing is assigned.

    toggle_show_default : `boolean`, optional
        Defines whether the options will be visible upon construction.

    toggle_show_visible : `boolean`, optional
        The visibility of the toggle button.
    """
    import IPython.html.widgets as ipywidgets
    # if image is not masked, then masked flag should be disabled
    if not channels_options_default['image_is_masked']:
        channels_options_default['masked_enabled'] = False

    # parse channels
    if isinstance(channels_options_default['channels'], list):
        if len(channels_options_default['channels']) == 1:
            mode_default = 'Single'
            first_slider_default = channels_options_default['channels'][0]
            second_slider_default = channels_options_default['n_channels'] - 1
        else:
            mode_default = 'Multiple'
            first_slider_default = min(channels_options_default['channels'])
            second_slider_default = max(channels_options_default['channels'])
    elif channels_options_default['channels'] is None:
        mode_default = 'Single'
        first_slider_default = 0
        second_slider_default = channels_options_default['n_channels'] - 1
    else:
        mode_default = 'Single'
        first_slider_default = channels_options_default['channels']
        second_slider_default = channels_options_default['n_channels'] - 1

    # Create all necessary widgets
    # If single channel, disable all options apart from masked
    but = ipywidgets.ToggleButtonWidget(description='Channels Options',
                                        value=toggle_show_default,
                                        visible=toggle_show_visible)
    mode = ipywidgets.RadioButtonsWidget(
        values=["Single", "Multiple"], value=mode_default, description='Mode:',
        visible=toggle_show_default,
        disabled=channels_options_default['n_channels'] == 1)
    masked = ipywidgets.CheckboxWidget(
        value=channels_options_default['masked_enabled'], description='Masked',
        visible=toggle_show_default and channels_options_default[
            'image_is_masked'])
    first_slider_wid = ipywidgets.IntSliderWidget(
        min=0, max=channels_options_default['n_channels'] - 1, step=1,
        value=first_slider_default, description='Channel',
        visible=toggle_show_default,
        disabled=channels_options_default['n_channels'] == 1)
    second_slider_wid = ipywidgets.IntSliderWidget(
        min=1, max=channels_options_default['n_channels'] - 1, step=1,
        value=second_slider_default, description='To',
        visible=mode_default == "Multiple")
    rgb_wid = ipywidgets.CheckboxWidget(
        value=channels_options_default['n_channels'] == 3 and
              channels_options_default['channels'] is None,
        description='RGB',
        visible=toggle_show_default and channels_options_default[
                                            'n_channels'] == 3)
    sum_wid = ipywidgets.CheckboxWidget(
        value=channels_options_default['sum_enabled'], description='Sum',
        visible=False, disabled=channels_options_default['n_channels'] == 1)
    glyph_wid = ipywidgets.CheckboxWidget(
        value=channels_options_default['glyph_enabled'], description='Glyph',
        visible=False, disabled=channels_options_default['n_channels'] == 1)
    glyph_block_size = ipywidgets.BoundedIntTextWidget(
        description='Block size', min=1, max=25,
        value=channels_options_default['glyph_block_size'], visible=False,
        disabled=channels_options_default['n_channels'] == 1)
    glyph_use_negative = ipywidgets.CheckboxWidget(
        description='Negative values',
        value=channels_options_default['glyph_use_negative'], visible=False,
        disabled=channels_options_default['n_channels'] == 1)

    # Group widgets
    glyph_options = ipywidgets.ContainerWidget(children=[glyph_block_size,
                                                         glyph_use_negative])
    glyph_all = ipywidgets.ContainerWidget(children=[glyph_wid, glyph_options])
    multiple_checkboxes = ipywidgets.ContainerWidget(
        children=[sum_wid, glyph_all, rgb_wid])
    sliders = ipywidgets.ContainerWidget(
        children=[first_slider_wid, second_slider_wid])
    all_but_radiobuttons = ipywidgets.ContainerWidget(
        children=[sliders, multiple_checkboxes])
    mode_and_masked = ipywidgets.ContainerWidget(children=[mode, masked])
    all_but_toggle = ipywidgets.ContainerWidget(children=[mode_and_masked,
                                                          all_but_radiobuttons])

    # Widget container
    channel_options_wid = ipywidgets.ContainerWidget(
        children=[but, all_but_toggle])

    # Initialize output variables
    channel_options_wid.selected_values = channels_options_default

    # Define mode visibility
    def mode_selection_fun(name, value):
        if value == 'Single':
            first_slider_wid.description = 'Channel'
            first_slider_wid.min = 0
            first_slider_wid.max = \
                channel_options_wid.selected_values['n_channels'] - 1
            second_slider_wid.visible = False
            sum_wid.visible = False
            sum_wid.value = False
            glyph_wid.visible = False
            glyph_wid.value = False
            glyph_block_size.visible = False
            glyph_block_size.value = '3'
            glyph_use_negative.visible = False
            glyph_use_negative.value = False
            rgb_wid.visible = \
                channel_options_wid.selected_values['n_channels'] == 3
            rgb_wid.value = \
                channel_options_wid.selected_values['n_channels'] == 3
        else:
            first_slider_wid.description = 'From'
            first_slider_wid.min = 0
            first_slider_wid.max = \
                channel_options_wid.selected_values['n_channels'] - 1
            second_slider_wid.min = 0
            second_slider_wid.max = \
                channel_options_wid.selected_values['n_channels'] - 1
            second_slider_wid.value = first_slider_wid.value
            second_slider_wid.visible = True
            rgb_wid.visible = False
            rgb_wid.value = False
            sum_wid.visible = True
            sum_wid.value = False
            glyph_wid.visible = True
            glyph_wid.value = False
            glyph_block_size.visible = False
            glyph_block_size.value = '3'
            glyph_use_negative.visible = False
            glyph_use_negative.value = False

    mode_selection_fun('', mode_default)
    mode.on_trait_change(mode_selection_fun, 'value')

    # Define glyph visibility
    def glyph_options_visibility_fun(name, value):
        glyph_block_size.visible = glyph_wid.value
        glyph_use_negative.visible = glyph_wid.value
        if glyph_wid.value:
            sum_wid.value = False

    glyph_wid.on_trait_change(glyph_options_visibility_fun, 'value')
    glyph_wid.on_trait_change(glyph_options_visibility_fun, 'visible')

    # Define rgb functionality
    def rgb_fun(name, value):
        first_slider_wid.disabled = value

    rgb_wid.on_trait_change(rgb_fun, 'value')

    # Define sum functionality
    def sum_fun(name, value):
        if value:
            glyph_wid.value = False

    sum_wid.on_trait_change(sum_fun, 'value')

    # Define masked functionality
    def masked_fun(name, value):
        channel_options_wid.masked_enabled = value

    masked.on_trait_change(masked_fun, 'value')

    # Function that gets glyph/sum options
    def get_glyph_options(name, value):
        channel_options_wid.selected_values['glyph_enabled'] = glyph_wid.value
        channel_options_wid.selected_values['sum_enabled'] = sum_wid.value
        channel_options_wid.selected_values[
            'glyph_use_negative'] = glyph_use_negative.value
        channel_options_wid.selected_values[
            'glyph_block_size'] = glyph_block_size.value
        if channel_options_wid.selected_values['sum_enabled']:
            channel_options_wid.selected_values['glyph_block_size'] = 1

    glyph_wid.on_trait_change(get_glyph_options, 'value')
    sum_wid.on_trait_change(get_glyph_options, 'value')
    glyph_use_negative.on_trait_change(get_glyph_options, 'value')
    glyph_block_size.on_trait_change(get_glyph_options, 'value')
    mode.on_trait_change(get_glyph_options, 'value')

    # Define multiple channels sliders functionality
    def first_slider_val(name, value):
        if mode.value == 'Multiple' and value > second_slider_wid.value:
            first_slider_wid.value = second_slider_wid.value

    def second_slider_val(name, value):
        if mode.value == 'Multiple' and value < first_slider_wid.value:
            second_slider_wid.value = first_slider_wid.value

    def get_channels(name, value):
        if mode.value == "Single":
            if rgb_wid.value:
                channel_options_wid.selected_values['channels'] = None
            else:
                channel_options_wid.selected_values[
                    'channels'] = first_slider_wid.value
        else:
            channel_options_wid.selected_values['channels'] = range(
                first_slider_wid.value,
                second_slider_wid.value + 1)

    first_slider_wid.on_trait_change(first_slider_val, 'value')
    second_slider_wid.on_trait_change(second_slider_val, 'value')
    first_slider_wid.on_trait_change(get_channels, 'value')
    second_slider_wid.on_trait_change(get_channels, 'value')
    rgb_wid.on_trait_change(get_channels, 'value')
    mode.on_trait_change(get_channels, 'value')

    def get_masked(name, value):
        channel_options_wid.selected_values['masked_enabled'] = value

    masked.on_trait_change(get_masked, 'value')

    # Toggle button function
    def toggle_image_options(name, value):
        mode.visible = value
        if value:
            masked.visible = channel_options_wid.selected_values[
                'image_is_masked']
            if mode.value == 'Single':
                first_slider_wid.visible = True
                visible = channel_options_wid.selected_values['n_channels'] == 3
                rgb_wid.visible = visible
            else:
                first_slider_wid.visible = True
                second_slider_wid.visible = True
                sum_wid.visible = True
                glyph_wid.visible = True
        else:
            masked.visible = False
            first_slider_wid.visible = False
            second_slider_wid.visible = False
            rgb_wid.visible = False
            sum_wid.visible = False
            glyph_wid.visible = False
            glyph_options.children[0].visible = False
            glyph_options.children[1].visible = False

    but.on_trait_change(toggle_image_options, 'value')

    # assign plot_function
    if plot_function is not None:
        # mode.on_trait_change(plot_function, 'value')
        masked.on_trait_change(plot_function, 'value')
        first_slider_wid.on_trait_change(plot_function, 'value')
        second_slider_wid.on_trait_change(plot_function, 'value')
        rgb_wid.on_trait_change(plot_function, 'value')
        sum_wid.on_trait_change(plot_function, 'value')
        glyph_wid.on_trait_change(plot_function, 'value')
        glyph_block_size.on_trait_change(plot_function, 'value')
        glyph_use_negative.on_trait_change(plot_function, 'value')

    return channel_options_wid


def format_channel_options(channel_options_wid, container_padding='6px',
                           container_margin='6px',
                           container_border='1px solid black',
                           toggle_button_font_weight='bold',
                           border_visible=True):
    r"""
    Function that corrects the align (style format) of a given channel_options
    widget. Usage example:
        channel_options_wid = channel_options()
        display(channel_options_wid)
        format_channel_options(channel_options_wid)

    Parameters
    ----------
    channel_options_wid :
        The widget object generated by the `channel_options()` function.

    container_padding : `str`, optional
        The padding around the widget, e.g. '6px'

    container_margin : `str`, optional
        The margin around the widget, e.g. '6px'

    container_border : `str`, optional
        The border around the widget, e.g. '1px solid black'

    toggle_button_font_weight : `str`
        The font weight of the toggle button, e.g. 'bold'

    border_visible : `boolean`, optional
        Defines whether to draw the border line around the widget.
    """
    # align glyph options
    channel_options_wid.children[1].children[1].children[1].children[
        1].children[1].remove_class('vbox')
    channel_options_wid.children[1].children[1].children[1].children[
        1].children[1].add_class('hbox')
    channel_options_wid.children[1].children[1].children[1].children[
        1].children[1].children[0].set_css('width', '0.8cm')

    # align sum and glyph checkboxes
    channel_options_wid.children[1].children[1].children[1].remove_class('vbox')
    channel_options_wid.children[1].children[1].children[1].add_class('hbox')

    # align radio buttons with the rest
    channel_options_wid.children[1].remove_class('vbox')
    channel_options_wid.children[1].add_class('hbox')
    channel_options_wid.children[1].add_class('align-start')

    # set toggle button font bold
    channel_options_wid.children[0].set_css('font-weight',
                                            toggle_button_font_weight)

    # margin and border around container widget
    channel_options_wid.set_css('padding', container_padding)
    channel_options_wid.set_css('margin', container_margin)
    if border_visible:
        channel_options_wid.set_css('border', container_border)


def update_channel_options(channel_options_wid, n_channels, image_is_masked,
                           masked_default=False):
    r"""
    Function that updates the state of a given channel_options widget if the
    image's number of channels or its masked flag has changed. Usage example:

    ::

        channel_options_wid = channel_options(n_channels=2,
                                              image_is_masked=True)
        display(channel_options_wid)
        format_channel_options(channel_options_wid)
        update_channel_options(channel_options_wid, n_channels=36,
                               image_is_masked=False)

    Parameters
    ----------
    channel_options_wid : widget
        The widget object generated by the `channel_options()` function.
    n_channels : `int`
        The number of channels.
    image_is_masked : `boolean`
        Flag that defines whether the image is an instance of :map:`MaskedImage`
        or subclass.
    masked_default : `boolean`, optional
        The value to be assigned at the masked checkbox.
    """
    # if image_is_masked flag has actually changed from the previous value
    if image_is_masked != channel_options_wid.selected_values['image_is_masked']:
        # change the channel_options output
        channel_options_wid.selected_values['image_is_masked'] = image_is_masked
        channel_options_wid.selected_values['masked_enabled'] = masked_default
        # set the masked checkbox state
        channel_options_wid.children[1].children[0].children[1].visible = \
            channel_options_wid.children[0].value and image_is_masked
        channel_options_wid.children[1].children[0].children[1].value = False

    # if n_channels are actually different from the previous value
    if n_channels != channel_options_wid.selected_values['n_channels']:
        # change the channel_options output
        channel_options_wid.selected_values['n_channels'] = n_channels
        channel_options_wid.selected_values['channels'] = 0
        # set the rgb checkbox state
        channel_options_wid.children[1].children[1].children[1].children[
            2].visible = \
            n_channels == 3 and channel_options_wid.children[0].value
        # set the channel options state (apart from the masked checkbox)
        if n_channels == 1:
            # set sliders max and min values
            channel_options_wid.children[1].children[1].children[0].children[
                0].max = 1
            channel_options_wid.children[1].children[1].children[0].children[
                1].max = 1
            channel_options_wid.children[1].children[1].children[0].children[
                0].min = 0
            channel_options_wid.children[1].children[1].children[0].children[
                1].min = 0
            # set sliders state
            channel_options_wid.children[1].children[1].children[0].children[
                0].disabled = True
            channel_options_wid.children[1].children[1].children[0].children[
                1].visible = False
            # set glyph/sum state
            channel_options_wid.children[1].children[1].children[1].children[
                0].disabled = True
            channel_options_wid.children[1].children[1].children[1].children[
                1].children[0].disabled = True
            channel_options_wid.children[1].children[1].children[1].children[
                1].children[1].children[0].disabled = True
            channel_options_wid.children[1].children[1].children[1].children[
                1].children[1].children[1].disabled = True
            # set mode state
            channel_options_wid.children[1].children[0].children[
                0].disabled = True
            # set mode and sliders values
            for k in range(4):
                if k == 0:
                    channel_options_wid.children[1].children[1].children[
                        1].children[2].value = False
                elif k == 1:
                    channel_options_wid.children[1].children[0].children[
                        0].value = "Single"
                elif k == 2:
                    channel_options_wid.children[1].children[1].children[
                        0].children[0].value = 0
                else:
                    channel_options_wid.children[1].children[1].children[
                        0].children[1].value = 0
        else:
            # set sliders max and min values
            channel_options_wid.children[1].children[1].children[0].children[
                0].max = n_channels - 1
            channel_options_wid.children[1].children[1].children[0].children[
                1].max = n_channels - 1
            channel_options_wid.children[1].children[1].children[0].children[
                0].min = 0
            channel_options_wid.children[1].children[1].children[0].children[
                1].min = 0
            # set sliders state
            channel_options_wid.children[1].children[1].children[0].children[
                0].disabled = False
            channel_options_wid.children[1].children[1].children[0].children[
                1].visible = False
            # set glyph/sum state
            channel_options_wid.children[1].children[1].children[1].children[
                0].disabled = False
            channel_options_wid.children[1].children[1].children[1].children[
                1].children[0].disabled = False
            channel_options_wid.children[1].children[1].children[1].children[
                1].children[1].children[0].disabled = False
            channel_options_wid.children[1].children[1].children[1].children[
                1].children[1].children[1].disabled = False
            # set mode state
            channel_options_wid.children[1].children[0].children[
                0].disabled = False
            # set mode and sliders values
            for k in range(4):
                if k == 0:
                    channel_options_wid.children[1].children[1].children[
                        1].children[2].value = n_channels == 3
                elif k == 1:
                    channel_options_wid.children[1].children[0].children[
                        0].value = "Single"
                elif k == 2:
                    channel_options_wid.children[1].children[1].children[
                        0].children[0].value = 0
                else:
                    channel_options_wid.children[1].children[1].children[
                        0].children[1].value = 0


def landmark_options(landmark_options_default, plot_function=None,
                     toggle_show_default=True, toggle_show_visible=True):
    r"""
    Creates a widget with Landmark Options. Specifically, it has:
        1) A checkbox that controls the landmarks' visibility.
        2) A drop down menu with the available landmark groups.
        3) Several toggle buttons with the group's available labels.
        4) A toggle button that controls the visibility of all the above, i.e.
           the landmark options.

    The structure of the widgets is the following:
        landmark_options_wid.children = [toggle_button, landmarks_checkbox,
                                         groups]
        groups.children = [group_drop_down_menu, labels]
        labels.children = [labels_text, labels_toggle_buttons]

    The returned widget saves the selected values in the following dictionary:
        landmark_options_wid.selected_values

    To fix the alignment within this widget, please refer to
    `format_landmark_options()` function.

    To update the state of this widget, please refer to
    `update_landmark_options()` function.

    Parameters
    ----------
    landmark_options_default : `dict`
        The default options. For example ::

            landmark_options_default = {'render_landmarks': True,
                                        'group_keys': ['PTS', 'ibug_face_68'],
                                        'labels_keys': [['all'], ['jaw', 'eye'],
                                        'group': 'PTS',
                                        'with_labels': ['all']}

    plot_function : `function` or None, optional
        The plot function that is executed when a widgets' value changes.
        If None, then nothing is assigned.
    toggle_show_default : `bool`, optional
        Defines whether the options will be visible upon construction.
    toggle_show_visible : `bool`, optional
        The visibility of the toggle button.
    """
    import IPython.html.widgets as ipywidgets
    # Create all necessary widgets
    but = ipywidgets.ToggleButtonWidget(description='Landmarks Options',
                                        value=toggle_show_default,
                                        visible=toggle_show_visible)
    landmarks = ipywidgets.CheckboxWidget(description='Render landmarks',
                                          value=landmark_options_default[
                                              'render_landmarks'])
    group = ipywidgets.DropdownWidget(
        values=landmark_options_default['group_keys'],
        description='Group')
    labels_toggles = [[ipywidgets.ToggleButtonWidget(description=k, value=True)
                       for k in s_keys]
                      for s_keys in landmark_options_default['labels_keys']]
    labels_text = ipywidgets.LatexWidget(value='Labels')
    labels = ipywidgets.ContainerWidget(children=labels_toggles[0])

    # Group widgets
    labels_and_text = ipywidgets.ContainerWidget(children=[labels_text, labels])
    group_wid = ipywidgets.ContainerWidget(children=[group, labels_and_text])

    # Widget container
    landmark_options_wid = ipywidgets.ContainerWidget(
        children=[but, landmarks, group_wid])

    # Initialize output variables
    landmark_options_wid.selected_values = landmark_options_default
    landmark_options_wid.selected_values['labels_toggles'] = labels_toggles
    landmark_options_wid.selected_values['group'] = \
        landmark_options_wid.selected_values['group_keys'][0]
    landmark_options_wid.selected_values['with_labels'] = \
        landmark_options_wid.selected_values['labels_keys'][0]

    # Disability control
    def landmarks_fun(name, value):
        # get landmarks_enabled value
        landmark_options_wid.selected_values['render_landmarks'] = value
        # disable group drop down menu
        group.disabled = not value
        # disable all labels toggles
        for s_keys in landmark_options_wid.selected_values['labels_toggles']:
            for k in s_keys:
                k.disabled = not value
        # if all currently selected labels toggles are False, set them all
        # to True
        all_values = [ww.value for ww in labels.children]
        if all(item is False for item in all_values):
            for ww in labels.children:
                ww.value = True

    landmarks_fun('', landmark_options_wid.selected_values['render_landmarks'])
    landmarks.on_trait_change(landmarks_fun, 'value')

    # Group drop down method
    def group_fun(name, value):
        # get group value
        landmark_options_wid.selected_values['group'] = value
        # assign the correct children to the labels toggles
        labels_toggles = landmark_options_wid.selected_values['labels_toggles']
        group_keys = landmark_options_wid.selected_values['group_keys']
        labels.children = labels_toggles[group_keys.index(value)]
        # get with_labels value
        landmark_options_wid.selected_values['with_labels'] = []
        for ww in labels.children:
            if ww.value:
                landmark_options_wid.selected_values['with_labels'].append(
                    str(ww.description))
        # assign plot_function to all enabled labels
        if plot_function is not None:
            for w in labels.children:
                w.on_trait_change(plot_function, 'value')

    group.on_trait_change(group_fun, 'value')

    # Labels function
    def labels_fun(name, value):
        # if all labels toggles are False, set landmarks checkbox to False
        all_values = [ww.value for ww in labels.children]
        if all(item is False for item in all_values):
            landmarks.value = False
        # get with_labels value
        landmark_options_wid.selected_values['with_labels'] = []
        for ww in labels.children:
            if ww.value:
                landmark_options_wid.selected_values['with_labels'].append(
                    str(ww.description))

    # assign labels_fun to all labels toggles (even hidden ones)
    for s_group in landmark_options_wid.selected_values['labels_toggles']:
        for w in s_group:
            w.on_trait_change(labels_fun, 'value')

    # Toggle button function
    def show_options(name, value):
        group_wid.visible = value
        landmarks.visible = value

    show_options('', toggle_show_default)
    but.on_trait_change(show_options, 'value')

    # assign plot_function
    if plot_function is not None:
        # assign plot_function to landmarks checkbox and group drop down menu
        landmarks.on_trait_change(plot_function, 'value')
        group.on_trait_change(plot_function, 'value')
        # assign plot_function to all currently active labels toggles
        for w in labels.children:
            w.on_trait_change(plot_function, 'value')

    return landmark_options_wid


def format_landmark_options(landmark_options_wid, container_padding='6px',
                            container_margin='6px',
                            container_border='1px solid black',
                            toggle_button_font_weight='bold',
                            border_visible=True):
    r"""
    Function that corrects the align (style format) of a given landmark_options
    widget. Usage example:
        landmark_options_wid = landmark_options()
        display(landmark_options_wid)
        format_landmark_options(landmark_options_wid)

    Parameters
    ----------
    landmark_options_wid :
        The widget object generated by the `landmark_options()` function.

    container_padding : `str`, optional
        The padding around the widget, e.g. '6px'

    container_margin : `str`, optional
        The margin around the widget, e.g. '6px'

    container_border : `str`, optional
        The border around the widget, e.g. '1px solid black'

    toggle_button_font_weight : `str`
        The font weight of the toggle button, e.g. 'bold'

    border_visible : `boolean`, optional
        Defines whether to draw the border line around the widget.
    """
    # align labels toggle buttons
    landmark_options_wid.children[2].children[1].children[1].remove_class('vbox')
    landmark_options_wid.children[2].children[1].children[1].add_class('hbox')

    # align labels buttons with text
    landmark_options_wid.children[2].children[1].children[0].set_css(
        'margin-right', '5px')
    landmark_options_wid.children[2].children[1].remove_class('vbox')
    landmark_options_wid.children[2].children[1].add_class('hbox')
    landmark_options_wid.children[2].children[1].add_class('align-center')

    # align group drop down menu with labels toggle buttons
    landmark_options_wid.children[2].children[1].set_css('margin-top', '10px')
    landmark_options_wid.children[2].add_class('align-start')

    # set toggle button font bold
    landmark_options_wid.children[0].set_css('font-weight',
                                             toggle_button_font_weight)

    # margin and border around container widget
    landmark_options_wid.set_css('padding', container_padding)
    landmark_options_wid.set_css('margin', container_margin)
    if border_visible:
        landmark_options_wid.set_css('border', container_border)


def update_landmark_options(landmark_options_wid, group_keys, labels_keys,
                            plot_function):
    r"""
    Function that updates the state of a given landmark_options widget if the
    group or label keys of an image has changed. Usage example:
        landmark_options_default = {'render_landmarks': True,
                                    'group_keys': ['PTS', 'ibug_face_68'],
                                    'labels_keys': [['all'], ['jaw', 'nose'],
                                    'group': 'PTS',
                                    'with_labels': ['all']}
        landmark_options_wid = landmark_options(landmark_options_default)
        display(landmark_options_wid)
        format_landmark_options(landmark_options_wid)
        update_landmark_options(landmark_options_wid,
                                group_keys=['group3'],
                                labels_keys=['label31', 'label32', 'label33'])
        format_landmark_options(landmark_options_wid)

    Note that the `format_landmark_options()` function needs to be called again
    after the `update_landmark_options()` function.

    Parameters
    ----------
    landmark_options_wid :
        The widget object generated by the `landmark_options()` function.

    group_keys : `list` of `str`
        A list of the available landmark groups.

    labels_keys : `list` of `list` of `str`
        A list of lists of each landmark group's labels.

    plot_function : `function` or None
        The plot function that is executed when a widgets' value changes.
        If None, then nothing is assigned.
    """
    import IPython.html.widgets as ipywidgets
    # check if the new group_keys and labels_keys are the same as the old
    # ones
    if not _compare_groups_and_labels(
            group_keys, labels_keys,
            landmark_options_wid.selected_values['group_keys'],
            landmark_options_wid.selected_values['labels_keys']):
        # Create all necessary widgets
        group = ipywidgets.DropdownWidget(values=group_keys,
                                          description='Group')
        labels_toggles = [
            [ipywidgets.ToggleButtonWidget(description=k, value=True)
             for k in s_keys] for s_keys in labels_keys]

        # Group widgets
        landmark_options_wid.children[2].children[1].children[1]. \
            children = labels_toggles[0]
        labels = landmark_options_wid.children[2].children[1]
        cont = ipywidgets.ContainerWidget(children=[group, labels])
        landmark_options_wid.children = [landmark_options_wid.children[0],
                                         landmark_options_wid.children[1],
                                         cont]

        # Initialize output variables
        landmark_options_wid.selected_values['group_keys'] = group_keys
        landmark_options_wid.selected_values['labels_keys'] = labels_keys
        landmark_options_wid.selected_values['labels_toggles'] = labels_toggles
        landmark_options_wid.selected_values['group'] = group_keys[0]
        landmark_options_wid.selected_values['with_labels'] = labels_keys[0]

        # Disability control
        def landmarks_fun(name, value):
            # get landmarks_enabled value
            landmark_options_wid.selected_values['render_landmarks'] = value
            # disable group drop down menu
            group.disabled = not value
            # disable all labels toggles
            for s_key in landmark_options_wid.selected_values['labels_toggles']:
                for k in s_key:
                    k.disabled = not value
            # if all currently selected labels toggles are False, set them all
            # to True
            children = (landmark_options_wid.children[2].children[1]
                                            .children[1].children)
            all_values = [ww.value for ww in children]
            if all(item is False for item in all_values):
                children = (landmark_options_wid.children[2].children[1]
                                                .children[1].children)
                for ww in children:
                    ww.value = True

        landmark_options_wid.children[1].on_trait_change(landmarks_fun, 'value')
        landmarks_fun('',
                      landmark_options_wid.selected_values['render_landmarks'])

        # Group drop down method
        def group_fun(name, value):
            # get group value
            landmark_options_wid.selected_values['group'] = value
            # assign the correct children to the labels toggles
            landmark_options_wid.children[2].children[1].children[1]. \
                children = \
            landmark_options_wid.selected_values['labels_toggles'][
                landmark_options_wid.selected_values['group_keys'].index(value)]
            # get with_labels value
            landmark_options_wid.selected_values['with_labels'] = []
            for ww in landmark_options_wid.children[2].children[1].children[
                1].children:
                if ww.value:
                    landmark_options_wid.selected_values['with_labels'].append(
                        str(ww.description))
            # assign plot_function to all enabled labels
            if plot_function is not None:
                children = (landmark_options_wid.children[2].children[1]
                                                .children[1].children)
                for w in children:
                    w.on_trait_change(plot_function, 'value')

        group.on_trait_change(group_fun, 'value')

        # Labels function
        def labels_fun(name, value):
            # if all labels toggles are False, set landmarks checkbox to False
            all_values = [ww.value
                          for ww in
                          landmark_options_wid.children[2].children[1].children[
                              1].children]
            if all(item is False for item in all_values):
                landmark_options_wid.children[1].value = False
            # get with_labels value
            landmark_options_wid.selected_values['with_labels'] = []
            for ww in landmark_options_wid.children[2].children[1].children[1]. \
                    children:
                if ww.value:
                    landmark_options_wid.selected_values['with_labels'].append(
                        str(ww.description))

        # assign labels_fun to all labels toggles (even hidden ones)
        for s_group in labels_toggles:
            for w in s_group:
                w.on_trait_change(labels_fun, 'value')

        # assign plot_function
        if plot_function is not None:
            # assign plot_function to landmarks checkbox, legend
            # checkbox and group drop down menu
            group.on_trait_change(plot_function, 'value')
            # assign plot_function to all currently active labels toggles
            for w in labels_toggles[0]:
                w.on_trait_change(plot_function, 'value')

        # Toggle button function
        def show_options(name, value):
            landmark_options_wid.children[1].visible = value
            landmark_options_wid.children[2].visible = value

        show_options('', landmark_options_wid.children[0].value)
        landmark_options_wid.children[0].on_trait_change(show_options, 'value')

        # If there is only one group with value ' ', this means that the image
        # didn't have any landmarks. So disable the show_landmarks checkbox.
        if len(group_keys) == 1 and group_keys[0] == ' ':
            # No landmarks are provided. So disable the show landmarks checkbox
            landmark_options_wid.children[1].value = False
            landmark_options_wid.children[1].disabled = True
        else:
            if landmark_options_wid.children[1].disabled:
                landmark_options_wid.children[1].disabled = False
                landmark_options_wid.children[1].value = True


def info_print(n_bullets, toggle_show_default=True, toggle_show_visible=True):
    r"""
    Creates a widget that can print information. Specifically, it has:
        1) A latex widget where user can write the info text in latex format.
        2) A toggle button that controls the visibility of all the above, i.e.
           the info printing.

    The structure of the widgets is the following:
        info_wid.children = [toggle_button, text_widget]

    Parameters
    ----------
    toggle_show_default : `boolean`, optional
        Defines whether the info will be visible upon construction.

    toggle_show_visible : `boolean`, optional
        The visibility of the toggle button.
    """
    import IPython.html.widgets as ipywidgets
    # Create toggle button
    but = ipywidgets.ToggleButtonWidget(description='Info',
                                        value=toggle_show_default,
                                        visible=toggle_show_visible)

    # Create text widget
    children = [ipywidgets.LatexWidget(value="> menpo")
                for _ in xrange(n_bullets)]
    text_wid = ipywidgets.ContainerWidget(children=children)

    # Toggle button function
    def show_options(name, value):
        text_wid.visible = value

    show_options('', toggle_show_default)
    but.on_trait_change(show_options, 'value')

    # Group widgets
    info_wid = ipywidgets.ContainerWidget(children=[but, text_wid])

    return info_wid


def format_info_print(info_wid, font_size_in_pt='9pt', container_padding='6px',
                      container_margin='6px',
                      container_border='1px solid black',
                      toggle_button_font_weight='bold',
                      border_visible=True):
    r"""
    Function that corrects the align (style format) of a given info widget.
    Usage example:
        info_wid = info_print()
        display(info_wid)
        format_info_print(info_wid)

    Parameters
    ----------
    info_wid :
        The widget object generated by the `info_print()` function.

    font_size_in_pt : `str`, optional
        The font size of the latex text, e.g. '9pt'

    container_padding : `str`, optional
        The padding around the widget, e.g. '6px'

    container_margin : `str`, optional
        The margin around the widget, e.g. '6px'

    container_border : `str`, optional
        The border around the widget, e.g. '1px solid black'

    toggle_button_font_weight : `str`
        The font weight of the toggle button, e.g. 'bold'

    border_visible : `boolean`, optional
        Defines whether to draw the border line around the widget.
    """
    # text widget formatting
    info_wid.children[1].set_css({'border': container_border,
                                  'padding': '4px',
                                  'margin-top': '1px'})

    # set font size
    for w in info_wid.children[1].children:
        w.set_css({'font-size': font_size_in_pt})

    # set toggle button font bold
    info_wid.children[0].set_css('font-weight', toggle_button_font_weight)

    # margin and border around container widget
    info_wid.set_css('padding', container_padding)
    info_wid.set_css('margin', container_margin)
    if border_visible:
        info_wid.set_css('border', container_border)


def animation_options(index_selection_default, plot_function=None,
                      update_function=None, index_description='Image Number',
                      index_minus_description='-', index_plus_description='+',
                      index_style='buttons', index_text_editable=True,
                      loop_default=False, interval_default=0.5,
                      toggle_show_title='Image Options',
                      toggle_show_default=True, toggle_show_visible=True):
    r"""
    Creates a widget for selecting an index and creating animations.
    Specifically, it has:
        1) An index selection widget. It can either be a slider or +/- buttons.
        2) A play toggle button.
        3) A stop toggle button.
        4) An options toggle button.
        If the options toggle is pressed, the following appear:
        5) An interval text area.
        6) A loop check box.

    The structure of the widget is the following:
        animation_options_wid.children = [toggle_button, options]
        options.children = [index_selection, animation]
        if index_style == 'buttons':
            index_selection.children = [title, minus_button, index_text,
                                        plus_button] (index_selection_buttons())
        elif index_style == 'slider':
            index_selection = index_slider (index_selection_slider())
        animation.children = [buttons, animation_options]
        buttons.children = [play_button, stop_button, play_options_button]
        animation_options.children = [interval_text, loop_checkbox]

    The returned widget saves the selected values in the following dictionary:
        animation_options_wid.selected_values
        animation_options_wid.index_style

    To fix the alignment within this widget please refer to
    `format_animation_options()` function.

    To update the state of this widget, please refer to
    `update_animation_options()` function.

    Parameters
    ----------
    index_selection_default : `dict`
        The dictionary with the default options. For example:
            index_selection_default = {'min':0,
                                       'max':100,
                                       'step':1,
                                       'index':10}

    plot_function : `function` or None, optional
        The plot function that is executed when the index value changes.
        If None, then nothing is assigned.

    update_function : `function` or None, optional
        The update function that is executed when the index value changes.
        If None, then nothing is assigned.

    index_description : `str`, optional
        The title of the index widget.

    index_minus_description : `str`, optional
        The title of the button that decreases the index.

    index_plus_description : `str`, optional
        The title of the button that increases the index.

    index_style : {``buttons`` or ``slider``}, optional
        If 'buttons', then 'index_selection_buttons()' is called.
        If 'slider', then 'index_selection_slider()' is called.

    index_text_editable : `boolean`, optional
        Flag that determines whether the index text will be editable.

    loop_default : `boolean`, optional
        If True, the animation makes loop.
        If False, the animation stops when reaching the index_max_value.

    interval_default : `float`, optional
        The interval between the animation frames.

    toggle_show_title : `str`, optional
        The title of the toggle button.

    toggle_show_default : `boolean`, optional
        Defines whether the options will be visible upon construction.

    toggle_show_visible : `boolean`, optional
        The visibility of the toggle button.
    """
    from time import sleep
    from IPython import get_ipython
    import IPython.html.widgets as ipywidgets

    # get the kernel to use it later in order to make sure that the widgets'
    # traits changes are passed during a while-loop
    kernel = get_ipython().kernel

    # Create index widget
    if index_style == 'slider':
        index_wid = index_selection_slider(index_selection_default,
                                           plot_function=plot_function,
                                           update_function=update_function,
                                           description=index_description)
    elif index_style == 'buttons':
        index_wid = index_selection_buttons(
            index_selection_default, plot_function=plot_function,
            update_function=update_function, description=index_description,
            minus_description=index_minus_description,
            plus_description=index_plus_description, loop=loop_default,
            text_editable=index_text_editable)

    # Create other widgets
    but = ipywidgets.ToggleButtonWidget(description=toggle_show_title,
                                        value=toggle_show_default,
                                        visible=toggle_show_visible)
    play_but = ipywidgets.ToggleButtonWidget(description='Play >', value=False)
    stop_but = ipywidgets.ToggleButtonWidget(description='Stop', value=True,
                                             disabled=True)
    play_options = ipywidgets.ToggleButtonWidget(description='Options',
                                                 value=False)
    loop = ipywidgets.CheckboxWidget(description='Loop', value=loop_default,
                                     visible=False)
    interval = ipywidgets.FloatTextWidget(description='Interval (sec)',
                                          value=interval_default, visible=False)

    # Widget container
    tmp_options = ipywidgets.ContainerWidget(children=[interval, loop])
    buttons = ipywidgets.ContainerWidget(
        children=[play_but, stop_but, play_options])
    animation = ipywidgets.ContainerWidget(children=[buttons, tmp_options])
    cont = ipywidgets.ContainerWidget(children=[index_wid, animation])
    animation_options_wid = ipywidgets.ContainerWidget(children=[but, cont])

    # Initialize variables
    animation_options_wid.selected_values = index_selection_default
    animation_options_wid.index_style = index_style

    # Play button pressed
    def play_press(name, value):
        stop_but.value = not value
        play_but.disabled = value
        play_options.disabled = value
        if value:
            play_options.value = False
    play_but.on_trait_change(play_press, 'value')

    # Stop button pressed
    def stop_press(name, value):
        play_but.value = not value
        stop_but.disabled = value
        play_options.disabled = not value
    stop_but.on_trait_change(stop_press, 'value')

    # show animation options checkbox function
    def play_options_fun(name, value):
        interval.visible = value
        loop.visible = value
    play_options.on_trait_change(play_options_fun, 'value')

    # animation function
    def play_fun(name, value):
        if loop.value:
            # loop is enabled
            i = animation_options_wid.selected_values['index']
            if i < animation_options_wid.selected_values['max']:
                i += animation_options_wid.selected_values['step']
            else:
                i = animation_options_wid.selected_values['min']

            ani_max_selected = animation_options_wid.selected_values['max']
            while i <= ani_max_selected and not stop_but.value:
                # update index value
                if index_style == 'slider':
                    index_wid.value = i
                else:
                    index_wid.children[2].value = i

                # Run IPython iteration.
                # This is the code that makes this operation non-blocking. This
                # will allow widget messages and callbacks to be processed.
                kernel.do_one_iteration()

                # update counter
                if i < animation_options_wid.selected_values['max']:
                    i += animation_options_wid.selected_values['step']
                else:
                    i = animation_options_wid.selected_values['min']

                # wait
                sleep(interval.value)
        else:
            # loop is disabled
            i = animation_options_wid.selected_values['index']
            i += animation_options_wid.selected_values['step']
            while (i <= animation_options_wid.selected_values['max'] and
                       not stop_but.value):
                # update value
                if index_style == 'slider':
                    index_wid.value = i
                else:
                    index_wid.children[2].value = i

                # Run IPython iteration.
                # This is the code that makes this operation non-blocking. This
                # will allow widget messages and callbacks to be processed.
                kernel.do_one_iteration()

                # update counter
                i += animation_options_wid.selected_values['step']

                # wait
                sleep(interval.value)
            if i > index_selection_default['max']:
                stop_but.value = True
    play_but.on_trait_change(play_fun, 'value')

    # Toggle button function
    def show_options(name, value):
        index_wid.visible = value
        buttons.visible = value
        interval.visible = False
        loop.visible = False
        if value:
            play_options.value = False
    show_options('', toggle_show_default)
    but.on_trait_change(show_options, 'value')

    return animation_options_wid


def format_animation_options(animation_options_wid, index_text_width='0.5cm',
                             container_padding='6px', container_margin='6px',
                             container_border='1px solid black',
                             toggle_button_font_weight='bold',
                             border_visible=True):
    r"""
    Function that corrects the align (style format) of a given animation_options
    widget. Usage example:
        animation_options_wid = animation_options()
        display(animation_options_wid)
        format_animation_options(animation_options_wid)

    Parameters
    ----------
    animation_options_wid :
        The widget object generated by the `animation_options()`
        function.

    index_text_width : `str`, optional
        The width of the index value text area.

    container_padding : `str`, optional
        The padding around the widget, e.g. '6px'

    container_margin : `str`, optional
        The margin around the widget, e.g. '6px'

    container_border : `str`, optional
        The border around the widget, e.g. '1px solid black'

    toggle_button_font_weight : `str`
        The font weight of the toggle button, e.g. 'bold'

    border_visible : `boolean`, optional
        Defines whether to draw the border line around the widget.
    """
    # format index widget
    format_index_selection(animation_options_wid.children[1].children[0],
                           text_width=index_text_width)

    # align play/stop button with animation options button
    animation_options_wid.children[1].children[1].children[0].remove_class(
        'vbox')
    animation_options_wid.children[1].children[1].children[0].add_class('hbox')
    animation_options_wid.children[1].children[1].add_class('align-end')

    # add margin on the right of the play button
    animation_options_wid.children[1].children[1].children[0].children[1]. \
        set_css('margin-right', container_margin)

    if animation_options_wid.index_style == 'slider':
        # align animation on the right of slider
        animation_options_wid.children[1].add_class('align-end')
    else:
        # align animation and index buttons
        animation_options_wid.children[1].remove_class('vbox')
        animation_options_wid.children[1].add_class('hbox')
        animation_options_wid.children[1].add_class('align-center')
        animation_options_wid.children[1].children[0].set_css('margin-right',
                                                              '1cm')

    # set interval width
    animation_options_wid.children[1].children[1].children[1].children[0]. \
        set_css('width', '20px')

    # set toggle button font bold
    animation_options_wid.children[0].set_css('font-weight',
                                              toggle_button_font_weight)

    # margin and border around container widget
    animation_options_wid.set_css('padding', container_padding)
    animation_options_wid.set_css('margin', container_margin)
    if border_visible:
        animation_options_wid.set_css('border', container_border)


def update_animation_options(animation_options_wid, index_selection_default,
                             plot_function=None, update_function=None):
    r"""
    Function that updates the state of a given animation_options widget if the
    index bounds have changed. Usage example:
        index_selection_default = {'min':0,
                                   'max':100,
                                   'step':1,
                                   'index':10}
        animation_options_wid = animation_options(index_selection_default)
        display(animation_options_wid)
        format_animation_options(animation_options_wid)
        index_selection_default = {'min':0,
                                   'max':10,
                                   'step':5,
                                   'index':5}
        update_animation_options(animation_options_wid, index_selection_default)

    Parameters
    ----------
    animation_options_wid :
        The widget object generated by either the `animation_options()`
        function.

    index_selection_default : `dict`
        The dictionary with the default options. For example:
            index_selection_default = {'min':0,
                                       'max':100,
                                       'step':1,
                                       'index':10}

    plot_function : `function` or None, optional
        The plot function that is executed when the index value changes.
        If None, then nothing is assigned.

    update_function : `function` or None, optional
        The update function that is executed when the index value changes.
        If None, then nothing is assigned.
    """
    update_index_selection(animation_options_wid.children[1].children[0],
                           index_selection_default,
                           plot_function=plot_function,
                           update_function=update_function)


def viewer_options(viewer_options_default, options_tabs, objects_names=None,
                   labels=None, plot_function=None, toggle_show_visible=True,
                   toggle_show_default=True):
    r"""
    Creates a widget with Viewer Options. Specifically, it has:
        1) A drop down menu for object selection.
        2) A tab widget with any of line, marker, numbers and feature options
        3) A toggle button that controls the visibility of all the above, i.e.
           the viewer options.

    The structure of the widgets is the following:
        viewer_options_wid.children = [toggle_button, options]
        options.children = [selection_menu, tab_options]
        tab_options.children = [line_options, marker_options,
                                numbers_options, figure_options, legend_options]

    The returned widget saves the selected values in the following dictionary:
        viewer_options_wid.selected_values

    To fix the alignment within this widget please refer to
    `format_viewer_options()` function.

    Parameters
    ----------
    viewer_options_default : list of `dict`
        A list of dictionaries with the initial selected viewer options per
        object. Example:

            lines_options = {'render_lines': True,
                             'line_width': 1,
                             'line_colour': ['b'],
                             'line_style': '-'}

            markers_options = {'render_markers':True,
                               'marker_size':20,
                               'marker_face_colour':['r'],
                               'marker_edge_colour':['k'],
                               'marker_style':'o',
                               'marker_edge_width':1}

            numbers_options = {'render_numbering': True,
                               'numbers_font_name': 'serif',
                               'numbers_font_size': 10,
                               'numbers_font_style': 'normal',
                               'numbers_font_weight': 'normal',
                               'numbers_font_colour': ['k'],
                               'numbers_horizontal_align': 'center',
                               'numbers_vertical_align': 'bottom'}

            legend_options = {'render_legend':True,
                              'legend_title':'',
                              'legend_font_name':'serif',
                              'legend_font_style':'normal',
                              'legend_font_size':10,
                              'legend_font_weight':'normal',
                              'legend_marker_scale':1.,
                              'legend_location':2,
                              'legend_bbox_to_anchor':(1.05, 1.),
                              'legend_border_axes_pad':1.,
                              'legend_n_columns':1,
                              'legend_horizontal_spacing':1.,
                              'legend_vertical_spacing':1.,
                              'legend_border':True,
                              'legend_border_padding':0.5,
                              'legend_shadow':False,
                              'legend_rounded_corners':True}

            figure_options = {'x_scale': 1.,
                              'y_scale': 1.,
                              'render_axes': True,
                              'axes_font_name': 'serif',
                              'axes_font_size': 10,
                              'axes_font_style': 'normal',
                              'axes_font_weight': 'normal',
                              'axes_x_limits': None,
                              'axes_y_limits': None}

            grid_options = {'render_grid': True,
                            'grid_line_style': '--',
                            'grid_line_width': 0.5}

            viewer_options_default = {'lines': lines_options,
                                      'markers': markers_options,
                                      'numbering': numbering_options,
                                      'legend': legend_options,
                                      'figure': figure_options,
                                      'grid': grid_options}

    options_tabs : `list` of `str`
        List that defines the ordering of the options tabs. It can take one of
        {``lines``, ``markers``, ``numbering``, ``figure_one``, ``figure_two``,
        ``legend``, ``grid``}

    objects_names : `list` of `str`, optional
        A list with the names of the objects that will be used in the selection
        dropdown menu. If None, then the names will have the form ``%d``.

    plot_function : `function` or None, optional
        The plot function that is executed when a widgets' value changes.
        If None, then nothing is assigned.

    toggle_show_default : `boolean`, optional
        Defines whether the options will be visible upon construction.

    toggle_show_visible : `boolean`, optional
        The visibility of the toggle button.
    """
    import IPython.html.widgets as ipywidgets
    # make sure that viewer_options_default is list even with one member
    if not isinstance(viewer_options_default, list):
        viewer_options_default = [viewer_options_default]

    # find number of objects
    n_objects = len(viewer_options_default)
    selection_visible = n_objects > 1

    # Create widgets
    # toggle button
    but = ipywidgets.ToggleButtonWidget(description='Viewer Options',
                                        value=toggle_show_default,
                                        visible=toggle_show_visible)

    # select object drop down menu
    objects_dict = OrderedDict()
    if objects_names is None:
        for k in range(n_objects):
            objects_dict[str(k)] = k
    else:
        for k, g in enumerate(objects_names):
            objects_dict[g] = k
    selection = ipywidgets.DropdownWidget(values=objects_dict, value=0,
                                          description='Select',
                                          visible=(selection_visible and
                                                   toggle_show_default))

    # options widgets
    options_widgets = []
    tab_titles = []
    if labels is None:
        labels = [str(j) for j in range(len(options_tabs))]
    for j, o in enumerate(options_tabs):
        if o == 'lines':
            options_widgets.append(
                line_options(viewer_options_default[0]['lines'],
                             toggle_show_visible=False,
                             toggle_show_default=True,
                             plot_function=plot_function,
                             show_checkbox_title='Render lines',
                             labels=labels[j]))
            tab_titles.append('Lines')
        elif o == 'markers':
            options_widgets.append(
                marker_options(viewer_options_default[0]['markers'],
                               toggle_show_visible=False,
                               toggle_show_default=True,
                               plot_function=plot_function,
                               show_checkbox_title='Render markers'))
            tab_titles.append('Markers')
        elif o == 'image':
            options_widgets.append(
                image_options(viewer_options_default[0]['image'],
                              toggle_show_visible=False,
                              toggle_show_default=True,
                              plot_function=plot_function))
            tab_titles.append('Image')
        elif o == 'numbering':
            options_widgets.append(
                numbering_options(viewer_options_default[0]['numbering'],
                                  toggle_show_visible=False,
                                  toggle_show_default=True,
                                  plot_function=plot_function,
                                  show_checkbox_title='Render numbering'))
            tab_titles.append('Numbering')
        elif o == 'figure_one':
            options_widgets.append(
                figure_options(viewer_options_default[0]['figure'],
                               plot_function=plot_function,
                               figure_scale_bounds=(0.1, 4),
                               figure_scale_step=0.1, figure_scale_visible=True,
                               axes_visible=True, toggle_show_default=True,
                               toggle_show_visible=False))
            tab_titles.append('Figure/Axes')
        elif o == 'figure_two':
            options_widgets.append(
                figure_options_two_scales(
                    viewer_options_default[0]['figure'],
                    plot_function=plot_function, coupled_default=False,
                    figure_scales_bounds=(0.1, 4), figure_scales_step=0.1,
                    figure_scales_visible=True, axes_visible=True,
                    toggle_show_default=True, toggle_show_visible=False))
            tab_titles.append('Figure/Axes')
        elif o == 'legend':
            options_widgets.append(
                legend_options(viewer_options_default[0]['legend'],
                               toggle_show_visible=False,
                               toggle_show_default=True,
                               plot_function=plot_function,
                               show_checkbox_title='Render legend'))
            tab_titles.append('Legend')
        elif o == 'grid':
            options_widgets.append(
                grid_options(viewer_options_default[0]['grid'],
                             toggle_show_visible=False,
                             toggle_show_default=True,
                             plot_function=plot_function,
                             show_checkbox_title='Render grid'))
            tab_titles.append('Grid')
    options = ipywidgets.TabWidget(children=options_widgets)

    # Final widget
    all_options = ipywidgets.ContainerWidget(children=[selection, options])
    viewer_options_wid = ipywidgets.ContainerWidget(children=[but, all_options])

    # save tab titles and options str to widget in order to be passed to the
    # format function
    viewer_options_wid.tab_titles = tab_titles
    viewer_options_wid.options_tabs = options_tabs

    # Assign output list of dicts
    viewer_options_wid.selected_values = viewer_options_default

    # Update widgets' state
    def update_widgets(name, value):
        for i, tab in enumerate(options_tabs):
            if tab == 'lines':
                update_line_options(
                    options_widgets[i],
                    viewer_options_default[value]['lines'],
                    labels=labels[value])
            elif tab == 'markers':
                update_marker_options(
                    options_widgets[i],
                    viewer_options_default[value]['markers'])
            elif tab == 'image':
                update_image_options(
                    options_widgets[i],
                    viewer_options_default[value]['image'])
            elif tab == 'numbering':
                update_numbering_options(
                    options_widgets[i],
                    viewer_options_default[value]['numbering'])
            elif tab == 'figure_one':
                update_figure_options(
                    options_widgets[i],
                    viewer_options_default[value]['figure'])
            elif tab == 'figure_two':
                update_figure_options_two_scales(
                    options_widgets[i],
                    viewer_options_default[value]['figure'])
            elif tab == 'legend':
                update_legend_options(
                    options_widgets[i],
                    viewer_options_default[value]['legend'])
            elif tab == 'grid':
                update_grid_options(
                    options_widgets[i],
                    viewer_options_default[value]['grid'])
    selection.on_trait_change(update_widgets, 'value')

    # Toggle button function
    def toggle_fun(name, value):
        selection.visible = value and selection_visible
        options.visible = value
    toggle_fun('', toggle_show_default)
    but.on_trait_change(toggle_fun, 'value')

    return viewer_options_wid


def format_viewer_options(viewer_options_wid, container_padding='6px',
                          container_margin='6px',
                          container_border='1px solid black',
                          toggle_button_font_weight='bold',
                          border_visible=False, suboptions_border_visible=True):
    r"""
    Function that corrects the align (style format) of a given
    viewer_options widget. Usage example:
        viewer_options_wid = viewer_options(default_options)
        display(viewer_options_wid)
        format_viewer_options(viewer_options_wid)

    Parameters
    ----------
    viewer_options_wid :
        The widget object generated by the `viewer_options()` function.

    container_padding : `str`, optional
        The padding around the widget, e.g. '6px'

    container_margin : `str`, optional
        The margin around the widget, e.g. '6px'

    container_border : `str`, optional
        The border around the widget, e.g. '1px solid black'

    toggle_button_font_weight : `str`
        The font weight of the toggle button, e.g. 'bold'

    border_visible : `boolean`, optional
        Defines whether to draw the border line around the widget.

    suboptions_border_visible : `boolean`, optional
        Defines whether to draw the border line around each of the sub options.
    """
    # format widgets
    for k, o in enumerate(viewer_options_wid.options_tabs):
        if o == 'lines':
            format_line_options(
                viewer_options_wid.children[1].children[1].children[k],
                suboptions_border_visible=suboptions_border_visible,
                border_visible=False)
        elif o == 'markers':
            format_marker_options(
                viewer_options_wid.children[1].children[1].children[k],
                suboptions_border_visible=suboptions_border_visible,
                border_visible=False)
        elif o == 'image':
            format_image_options(
                viewer_options_wid.children[1].children[1].children[k],
                border_visible=suboptions_border_visible)
        elif o == 'numbering':
            format_numbering_options(
                viewer_options_wid.children[1].children[1].children[k],
                suboptions_border_visible=suboptions_border_visible,
                border_visible=False)
        elif o == 'figure_one':
            format_figure_options(
                viewer_options_wid.children[1].children[1].children[k],
                border_visible=suboptions_border_visible)
        elif o == 'figure_two':
            format_figure_options_two_scales(
                viewer_options_wid.children[1].children[1].children[k],
                border_visible=suboptions_border_visible)
        elif o == 'legend':
            format_legend_options(
                viewer_options_wid.children[1].children[1].children[k],
                suboptions_border_visible=suboptions_border_visible,
                border_visible=False)
        elif o == 'grid':
            format_grid_options(
                viewer_options_wid.children[1].children[1].children[k],
                suboptions_border_visible=suboptions_border_visible,
                border_visible=False)

    # set titles
    for (k, tl) in enumerate(viewer_options_wid.tab_titles):
        viewer_options_wid.children[1].children[1].set_title(k, tl)

    # set toggle button font bold
    viewer_options_wid.children[0].set_css('font-weight',
                                           toggle_button_font_weight)

    # margin and border around container widget
    viewer_options_wid.set_css('padding', container_padding)
    viewer_options_wid.set_css('margin', container_margin)
    if border_visible:
        viewer_options_wid.set_css('border', container_border)


def update_viewer_options(viewer_options_wid, viewer_options_default,
                          labels=None):
    for k, o in enumerate(viewer_options_wid.options_tabs):
        if o == 'lines' and 'lines' in viewer_options_default:
            update_line_options(
                viewer_options_wid.children[1].children[1].children[k],
                viewer_options_default['lines'], labels=labels)
        elif o == 'markers' and 'markers' in viewer_options_default:
            update_marker_options(
                viewer_options_wid.children[1].children[1].children[k],
                viewer_options_default['markers'])
        elif o == 'image' and 'image' in viewer_options_default:
            update_image_options(
                viewer_options_wid.children[1].children[1].children[k],
                viewer_options_default['image'])
        elif o == 'numbering' and 'numbering' in viewer_options_default:
            update_numbering_options(
                viewer_options_wid.children[1].children[1].children[k],
                viewer_options_default['numbering'])
        elif o == 'figure_one' and 'figure' in viewer_options_default:
            update_figure_options(
                viewer_options_wid.children[1].children[1].children[k],
                viewer_options_default['figure'])
        elif o == 'figure_two' and 'figure' in viewer_options_default:
            update_figure_options(
                viewer_options_wid.children[1].children[1].children[k],
                viewer_options_default['figure'])
        elif o == 'legend' and 'legend' in viewer_options_default:
            update_legend_options(
                viewer_options_wid.children[1].children[1].children[k],
                viewer_options_default['legend'])
        elif o == 'grid' and 'grid' in viewer_options_default:
            update_grid_options(
                viewer_options_wid.children[1].children[1].children[k],
                viewer_options_default['grid'])


def save_figure_options(renderer, format_default='png', dpi_default=None,
                        orientation_default='portrait',
                        papertype_default='letter', transparent_default=False,
                        facecolour_default='w', edgecolour_default='w',
                        pad_inches_default=0.5, overwrite_default=False,
                        toggle_show_default=True, toggle_show_visible=True):
    r"""
    Creates a widget with Save Figure Options.

    The structure of the widgets is the following:
        save_figure_wid.children = [toggle_button, options, save_button]
        options.children = [path, page_setup, image_colour]
        path.children = [filename, format, papertype]
        page_setup.children = [orientation, dpi, pad_inches]
        image_colour.children = [facecolour, edgecolour, transparent]

    To fix the alignment within this widget please refer to
    `format_save_figure_options()` function.

    Parameters
    ----------
    figure_id : matplotlib.pyplot.Figure instance
        The handle of the figure to be saved.

    format_default : `str`, optional
        The default value of the format.

    dpi_default : `float`, optional
        The default value of the dpi.

    orientation_default : `str`, optional
        The default value of the orientation. 'portrait' or 'landscape'.

    papertype_default : `str`, optional
        The default value of the papertype.

    transparent_default : `boolean`, optional
        The default value of the transparency flag.

    facecolour_default : `str` or `list` of `float`, optional
        The default value of the facecolour.

    edgecolour_default : `str` or `list` of `float`, optional
        The default value of the edgecolour.

    pad_inches_default : `float`, optional
        The default value of the figure padding in inches.

    toggle_show_default : `boolean`, optional
        Defines whether the options will be visible upon construction.

    toggle_show_visible : `boolean`, optional
        The visibility of the toggle button.
    """
    import IPython.html.widgets as ipywidgets
    from os import getcwd
    from os.path import join, splitext

    # create widgets
    but = ipywidgets.ToggleButtonWidget(description='Save Figure',
                                        value=toggle_show_default,
                                        visible=toggle_show_visible)
    format_dict = OrderedDict()
    format_dict['png'] = 'png'
    format_dict['jpg'] = 'jpg'
    format_dict['pdf'] = 'pdf'
    format_dict['eps'] = 'eps'
    format_dict['postscript'] = 'ps'
    format_dict['svg'] = 'svg'
    format_wid = ipywidgets.SelectWidget(values=format_dict,
                                         value=format_default,
                                         description='Format')

    def papertype_visibility(name, value):
        papertype_wid.disabled = not value == 'ps'

    format_wid.on_trait_change(papertype_visibility, 'value')

    def set_extension(name, value):
        fileName, fileExtension = splitext(filename.value)
        filename.value = fileName + '.' + value

    format_wid.on_trait_change(set_extension, 'value')
    if dpi_default is None:
        dpi_default = 0
    dpi_wid = ipywidgets.FloatTextWidget(description='DPI', value=dpi_default)
    orientation_dict = OrderedDict()
    orientation_dict['portrait'] = 'portrait'
    orientation_dict['landscape'] = 'landscape'
    orientation_wid = ipywidgets.DropdownWidget(values=orientation_dict,
                                                value=orientation_default,
                                                description='Orientation')
    papertype_dict = OrderedDict()
    papertype_dict['letter'] = 'letter'
    papertype_dict['legal'] = 'legal'
    papertype_dict['executive'] = 'executive'
    papertype_dict['ledger'] = 'ledger'
    papertype_dict['a0'] = 'a0'
    papertype_dict['a1'] = 'a1'
    papertype_dict['a2'] = 'a2'
    papertype_dict['a3'] = 'a3'
    papertype_dict['a4'] = 'a4'
    papertype_dict['a5'] = 'a5'
    papertype_dict['a6'] = 'a6'
    papertype_dict['a7'] = 'a7'
    papertype_dict['a8'] = 'a8'
    papertype_dict['a9'] = 'a9'
    papertype_dict['a10'] = 'a10'
    papertype_dict['b0'] = 'b0'
    papertype_dict['b1'] = 'b1'
    papertype_dict['b2'] = 'b2'
    papertype_dict['b3'] = 'b3'
    papertype_dict['b4'] = 'b4'
    papertype_dict['b5'] = 'b5'
    papertype_dict['b6'] = 'b6'
    papertype_dict['b7'] = 'b7'
    papertype_dict['b8'] = 'b8'
    papertype_dict['b9'] = 'b9'
    papertype_dict['b10'] = 'b10'
    is_ps_type = not format_default == 'ps'
    papertype_wid = ipywidgets.DropdownWidget(values=papertype_dict,
                                              value=papertype_default,
                                              description='Paper type',
                                              disabled=is_ps_type)
    transparent_wid = ipywidgets.CheckboxWidget(description='Transparent',
                                                value=transparent_default)
    facecolour_wid = colour_selection([facecolour_default], title='Face colour')
    edgecolour_wid = colour_selection([edgecolour_default], title='Edge colour')
    pad_inches_wid = ipywidgets.FloatTextWidget(description='Pad (inch)',
                                                value=pad_inches_default)
    filename = ipywidgets.TextWidget(description='Path',
                                     value=join(getcwd(),
                                                'Untitled.' + format_default))
    overwrite = ipywidgets.CheckboxWidget(
        description='Overwrite if file exists',
        value=overwrite_default)
    error_str = ipywidgets.LatexWidget(value="")
    save_but = ipywidgets.ButtonWidget(description='Save')

    # create final widget
    path_wid = ipywidgets.ContainerWidget(
        children=[filename, format_wid, overwrite,
                  papertype_wid])
    page_wid = ipywidgets.ContainerWidget(children=[orientation_wid, dpi_wid,
                                                    pad_inches_wid])
    colour_wid = ipywidgets.ContainerWidget(
        children=[facecolour_wid, edgecolour_wid,
                  transparent_wid])
    options_wid = ipywidgets.TabWidget(
        children=[path_wid, page_wid, colour_wid])
    save_wid = ipywidgets.ContainerWidget(children=[save_but, error_str])
    save_figure_wid = ipywidgets.ContainerWidget(
        children=[but, options_wid, save_wid])

    # Assign renderer
    save_figure_wid.renderer = [renderer]

    # save function
    def save_function(name):
        # set save button state
        error_str.value = ''
        save_but.description = 'Saving...'
        save_but.disabled = True

        # save figure
        selected_dpi = dpi_wid.value
        if dpi_wid.value == 0:
            selected_dpi = None
        try:
            save_figure_wid.renderer[0].save_figure(
                filename=filename.value, dpi=selected_dpi,
                face_colour=facecolour_wid.selected_values['colour'][0],
                edge_colour=edgecolour_wid.selected_values['colour'][0],
                orientation=orientation_wid.value,
                paper_type=papertype_wid.value, format=format_wid.value,
                transparent=transparent_wid.value,
                pad_inches=pad_inches_wid.value, overwrite=overwrite.value)
            error_str.value = ''
        except ValueError as e:
            if (e.message == 'File already exists. Please set the overwrite '
                             'kwarg if you wish to overwrite the file.'):
                error_str.value = 'File exists! Select overwrite to replace.'
            else:
                error_str.value = e.message

        # set save button state
        save_but.description = 'Save'
        save_but.disabled = False
    save_but.on_click(save_function)

    # Toggle button function
    def show_options(name, value):
        options_wid.visible = value
        save_but.visible = value
    show_options('', toggle_show_default)
    but.on_trait_change(show_options, 'value')

    return save_figure_wid


def format_save_figure_options(save_figure_wid, container_padding='6px',
                               container_margin='6px',
                               container_border='1px solid black',
                               toggle_button_font_weight='bold',
                               tab_top_margin='0.3cm',
                               border_visible=True):
    r"""
    Function that corrects the align (style format) of a given
    save_figure_options widget. Usage example:
        save_figure_wid = save_figure_options()
        display(save_figure_wid)
        format_save_figure_options(save_figure_wid)

    Parameters
    ----------
    save_figure_wid :
        The widget object generated by the `save_figure_options()` function.

    container_padding : `str`, optional
        The padding around the widget, e.g. '6px'

    container_margin : `str`, optional
        The margin around the widget, e.g. '6px'

    tab_top_margin : `str`, optional
        The margin around the tab options' widget, e.g. '0.3cm'

    container_border : `str`, optional
        The border around the widget, e.g. '1px solid black'

    toggle_button_font_weight : `str`
        The font weight of the toggle button, e.g. 'bold'

    border_visible : `boolean`, optional
        Defines whether to draw the border line around the widget.
    """
    # add margin on top of tabs widget
    save_figure_wid.children[1].set_css('margin-top', tab_top_margin)

    # align path options to the right
    save_figure_wid.children[1].children[0].add_class('align-end')

    # align save button and error message horizontally
    save_figure_wid.children[2].remove_class('vbox')
    save_figure_wid.children[2].add_class('hbox')
    save_figure_wid.children[2].children[1].set_css({'margin-left': '0.5cm',
                                                     'background-color': 'red'})

    # set final tab titles
    tab_titles = ['Path', 'Page setup', 'Image colour']
    for (k, tl) in enumerate(tab_titles):
        save_figure_wid.children[1].set_title(k, tl)

    format_colour_selection(save_figure_wid.children[1].children[2].children[0])
    format_colour_selection(save_figure_wid.children[1].children[2].children[1])
    save_figure_wid.children[1].children[0].children[0].set_css('width', '6cm')
    save_figure_wid.children[1].children[0].children[1].set_css('width', '6cm')

    # set toggle button font bold
    save_figure_wid.children[0].set_css('font-weight',
                                        toggle_button_font_weight)

    # margin and border around container widget
    save_figure_wid.set_css('padding', container_padding)
    save_figure_wid.set_css('margin', container_margin)
    if border_visible:
        save_figure_wid.set_css('border', container_border)


def features_options(toggle_show_default=True, toggle_show_visible=True):
    r"""
    Creates a widget with Features Options.

    The structure of the widgets is the following:
        features_options_wid.children = [toggle_button, tab_options]
        tab_options.children = [features_radiobuttons, per_feature_options,
                                preview]
        per_feature_options.children = [hog_options, igo_options, lbp_options,
                                        daisy_options, no_options]
        preview.children = [input_size_text, lenna_image, output_size_text,
                            elapsed_time]

    To fix the alignment within this widget please refer to
    `format_features_options()` function.

    Parameters
    ----------
    toggle_show_default : `boolean`, optional
        Defines whether the options will be visible upon construction.

    toggle_show_visible : `boolean`, optional
        The visibility of the toggle button.
    """
    # import features methods and time
    import time
    from menpo.feature.features import hog, lbp, igo, es, daisy, gradient, no_op
    from menpo.image import Image
    import menpo.io as mio
    from menpo.visualize.image import glyph
    import IPython.html.widgets as ipywidgets

    # Toggle button that controls options' visibility
    but = ipywidgets.ToggleButtonWidget(description='Features Options',
                                        value=toggle_show_default,
                                        visible=toggle_show_visible)

    # feature type
    tmp = OrderedDict()
    tmp['HOG'] = hog
    tmp['IGO'] = igo
    tmp['ES'] = es
    tmp['Daisy'] = daisy
    tmp['LBP'] = lbp
    tmp['Gradient'] = gradient
    tmp['None'] = no_op
    feature = ipywidgets.RadioButtonsWidget(value=no_op, values=tmp,
                                            description='Feature type:')

    # feature-related options
    hog_options_wid = hog_options(toggle_show_default=True,
                                  toggle_show_visible=False)
    igo_options_wid = igo_options(toggle_show_default=True,
                                  toggle_show_visible=False)
    lbp_options_wid = lbp_options(toggle_show_default=True,
                                  toggle_show_visible=False)
    daisy_options_wid = daisy_options(toggle_show_default=True,
                                      toggle_show_visible=False)
    no_options_wid = ipywidgets.LatexWidget(value='No options available.')

    # load and rescale preview image (lenna)
    image = mio.import_builtin_asset.lenna_png()
    image.crop_to_landmarks_proportion_inplace(0.18)
    image = image.as_greyscale()

    # per feature options widget
    per_feature_options = ipywidgets.ContainerWidget(
        children=[hog_options_wid, igo_options_wid, lbp_options_wid,
                  daisy_options_wid, no_options_wid])

    # preview tab widget
    preview_img = ipywidgets.ImageWidget(value=_convert_image_to_bytes(image),
                                         visible=False)
    preview_input = ipywidgets.LatexWidget(
        value="Input: {}W x {}H x {}C".format(
            image.width, image.height, image.n_channels), visible=False)
    preview_output = ipywidgets.LatexWidget(value="")
    preview_time = ipywidgets.LatexWidget(value="")
    preview = ipywidgets.ContainerWidget(children=[preview_img, preview_input,
                                                   preview_output,
                                                   preview_time])

    # options tab widget
    all_options = ipywidgets.TabWidget(
        children=[feature, per_feature_options, preview])

    # Widget container
    features_options_wid = ipywidgets.ContainerWidget(
        children=[but, all_options])

    # Initialize output dictionary
    options = {}
    features_options_wid.function = partial(no_op, **options)
    features_options_wid.features_function = no_op
    features_options_wid.features_options = options

    # options visibility
    def per_feature_options_visibility(name, value):
        if value == hog:
            igo_options_wid.visible = False
            lbp_options_wid.visible = False
            daisy_options_wid.visible = False
            no_options_wid.visible = False
            hog_options_wid.visible = True
        elif value == igo:
            hog_options_wid.visible = False
            lbp_options_wid.visible = False
            daisy_options_wid.visible = False
            no_options_wid.visible = False
            igo_options_wid.visible = True
        elif value == lbp:
            hog_options_wid.visible = False
            igo_options_wid.visible = False
            daisy_options_wid.visible = False
            no_options_wid.visible = False
            lbp_options_wid.visible = True
        elif value == daisy:
            hog_options_wid.visible = False
            igo_options_wid.visible = False
            lbp_options_wid.visible = False
            no_options_wid.visible = False
            daisy_options_wid.visible = True
        else:
            hog_options_wid.visible = False
            igo_options_wid.visible = False
            lbp_options_wid.visible = False
            daisy_options_wid.visible = False
            no_options_wid.visible = True
            for name, f in tmp.iteritems():
                if f == value:
                    no_options_wid.value = "{}: No available " \
                                           "options.".format(name)
    feature.on_trait_change(per_feature_options_visibility, 'value')
    per_feature_options_visibility('', no_op)

    # get function
    def get_function(name, value):
        # get options
        if feature.value == hog:
            opts = hog_options_wid.options
        elif feature.value == igo:
            opts = igo_options_wid.options
        elif feature.value == lbp:
            opts = lbp_options_wid.options
        elif feature.value == daisy:
            opts = daisy_options_wid.options
        else:
            opts = {}
        # get features function closure
        func = partial(feature.value, **opts)
        # store function
        features_options_wid.function = func
        features_options_wid.features_function = value
        features_options_wid.features_options = opts
    feature.on_trait_change(get_function, 'value')
    all_options.on_trait_change(get_function, 'selected_index')

    # preview function
    def preview_function(name, old_value, value):
        if value == 2:
            # extracting features message
            for name, f in tmp.iteritems():
                if f == features_options_wid.function.func:
                    val1 = name
            preview_output.value = "Previewing {} features...".format(val1)
            preview_time.value = ""
            # extract feature and time it
            t = time.time()
            feat_image = features_options_wid.function(image)
            t = time.time() - t
            # store feature image shape and n_channels
            val2 = feat_image.width
            val3 = feat_image.height
            val4 = feat_image.n_channels
            # compute sum of feature image and normalize its pixels in range
            # (0, 1) because it is required by as_PILImage
            feat_image = glyph(feat_image, vectors_block_size=1,
                               use_negative=False)
            # feat_image = np.sum(feat_image.pixels, axis=2)
            feat_image = feat_image.pixels
            feat_image -= np.min(feat_image)
            feat_image /= np.max(feat_image)
            feat_image = Image(feat_image)
            # update preview
            preview_img.value = _convert_image_to_bytes(feat_image)
            preview_input.visible = True
            preview_img.visible = True
            # set info
            preview_output.value = "{}: {}W x {}H x {}C".format(val1, val2,
                                                                val3, val4)
            preview_time.value = "{0:.2f} secs elapsed".format(t)
        if old_value == 2:
            preview_input.visible = False
            preview_img.visible = False
    all_options.on_trait_change(preview_function, 'selected_index')

    # Toggle button function
    def toggle_options(name, value):
        all_options.visible = value
    but.on_trait_change(toggle_options, 'value')

    return features_options_wid


def format_features_options(features_options_wid, container_padding='6px',
                            container_margin='6px',
                            container_border='1px solid black',
                            toggle_button_font_weight='bold',
                            border_visible=True):
    r"""
    Function that corrects the align (style format) of a given features_options
    widget. Usage example:
        features_options_wid = features_options()
        display(features_options_wid)
        format_features_options(features_options_wid)

    Parameters
    ----------
    features_options_wid :
        The widget object generated by the `features_options()` function.

    container_padding : `str`, optional
        The padding around the widget, e.g. '6px'

    container_margin : `str`, optional
        The margin around the widget, e.g. '6px'

    tab_top_margin : `str`, optional
        The margin around the tab options' widget, e.g. '0.3cm'

    container_border : `str`, optional
        The border around the widget, e.g. '1px solid black'

    toggle_button_font_weight : `str`
        The font weight of the toggle button, e.g. 'bold'

    border_visible : `boolean`, optional
        Defines whether to draw the border line around the widget.
    """
    # format per feature options
    format_hog_options(features_options_wid.children[1].children[1].children[0],
                       border_visible=False)
    format_igo_options(features_options_wid.children[1].children[1].children[1],
                       border_visible=False)
    format_lbp_options(features_options_wid.children[1].children[1].children[2],
                       border_visible=False)
    format_daisy_options(
        features_options_wid.children[1].children[1].children[3],
        border_visible=False)

    # set final tab titles
    tab_titles = ['Feature', 'Options', 'Preview']
    for (k, tl) in enumerate(tab_titles):
        features_options_wid.children[1].set_title(k, tl)

    # set margin above tab widget
    features_options_wid.children[1].set_css('margin', '10px')

    # set toggle button font bold
    features_options_wid.children[0].set_css('font-weight',
                                             toggle_button_font_weight)

    # margin and border around container widget
    features_options_wid.set_css('padding', container_padding)
    features_options_wid.set_css('margin', container_margin)
    if border_visible:
        features_options_wid.set_css('border', container_border)


def _compare_groups_and_labels(groups1, labels1, groups2, labels2):
    r"""
    Function that compares two sets of landmarks groups and labels and returns
    Trues if they are identical else False.

    Parameters
    ----------
    group1 : `list` of `str`
        The first list of landmark groups.

    labels1 : `list` of `list` of `str`
        The first list of lists of each landmark group's labels.

    group2 : `list` of `str`
        The second list of landmark groups.

    labels2 : `list` of `list` of `str`
        The second list of lists of each landmark group's labels.
    """
    # function that compares two lists without taking into account the order
    def comp_lists(l1, l2):
        len_match = len(l1) == len(l2)
        return len_match and np.all([g1 == g2 for g1, g2 in zip(l1, l2)])

    # comparison of the given groups
    groups_same = comp_lists(groups1, groups2)

    # if groups are the same, compare the labels
    if groups_same:
        len_match = len(labels1) == len(labels2)
        return len_match and np.all([comp_lists(g1, g2)
                                     for g1, g2 in zip(labels1, labels2)])
    else:
        return False



import numpy as np
import menpo
from numpy.testing import assert_allclose
from menpo.image import BooleanImage, Image, MaskedImage
from menpo.transform import Affine
import menpo.io as mio

# do the import to generate the expected outputs
rgb_image = mio.import_builtin_asset('takeo.ppm')
gray_image = rgb_image.as_greyscale()
gray_template = gray_image.crop(np.array([70, 30]),
                                np.array([169, 129]))
rgb_template = rgb_image.crop(np.array([70, 30]),
                              np.array([169, 129]))
template_mask = BooleanImage.blank(gray_template.shape)

initial_params = np.array([0, 0, 0, 0, 70, 30])
row_indices, col_indices = np.meshgrid(np.arange(50, 100), np.arange(50, 100),
                                       indexing='ij')
row_indices, col_indices = row_indices.flatten(), col_indices.flatten()
multi_expected = rgb_image.crop([50, 50],
                                [100, 100]).pixels.flatten()


def test_warp_gray():
    rgb_image = mio.import_builtin_asset('takeo.ppm')
    gray_image = rgb_image.as_greyscale()
    target_transform = Affine.identity(2).from_vector(initial_params)
    warped_im = gray_image.warp_to_mask(template_mask, target_transform)

    assert(warped_im.shape == gray_template.shape)
    assert_allclose(warped_im.pixels, gray_template.pixels)


def test_warp_multi():
    rgb_image = mio.import_builtin_asset('takeo.ppm')
    target_transform = Affine.identity(2).from_vector(initial_params)
    warped_im = rgb_image.warp_to_mask(template_mask, target_transform)

    assert(warped_im.shape == rgb_template.shape)
    assert_allclose(warped_im.pixels, rgb_template.pixels)


def test_warp_to_mask_boolean():
    b = BooleanImage.blank((10, 10))
    b.pixels[:, :5] = False
    template_mask = BooleanImage.blank((10, 10))
    template_mask.pixels[:5, :] = False
    t = Affine.identity(2)
    warped_mask = b.warp_to_mask(template_mask, t)
    assert(type(warped_mask) == BooleanImage)
    result = template_mask.pixels.copy()
    result[:, :5] = False
    assert(np.all(result == warped_mask.pixels))


def test_warp_to_mask_image():
    img = Image.blank((10, 10), n_channels=2)
    img.pixels[:, :5, :] = 0.5
    template_mask = BooleanImage.blank((10, 10))
    template_mask.pixels[5:, :] = False
    t = Affine.identity(2)
    warped_img = img.warp_to_mask(template_mask, t)
    assert(type(warped_img) == MaskedImage)
    result = Image.blank((10, 10), n_channels=2).pixels
    result[:5, :5, :] = 0.5
    assert(np.all(result == warped_img.pixels))


def test_warp_to_mask_masked_image():
    mask = BooleanImage.blank((10, 10))
    # make a funny mask on the original image
    mask.pixels[2:, :] = False
    img = MaskedImage.blank((10, 10), n_channels=2, mask=mask)
    img.pixels[...] = 2.5
    template_mask = BooleanImage.blank((10, 10), fill=False)
    template_mask.pixels[:5, :5] = True
    t = Affine.identity(2)
    warped_img = img.warp_to_mask(template_mask, t)
    assert(type(warped_img) == MaskedImage)
    result = Image.blank((10, 10), n_channels=2).pixels
    result[:5, :5, :] = 2.5
    result_mask = BooleanImage.blank((10, 10), fill=False).pixels
    result_mask[:2, :5] = True
    assert(warped_img.n_true_pixels() == 10)
    assert(np.all(result == warped_img.pixels))
    assert(np.all(result_mask == warped_img.mask.pixels))


def test_warp_to_shape_equal_warp_to_mask():
    r = menpo.transform.UniformScale(2.0, n_dims=2)
    b = mio.import_builtin_asset('breakingbad.jpg')
    m_shape = b.warp_to_shape((540, 960), r)
    m_mask = b.warp_to_mask(menpo.image.BooleanImage.blank((540, 960)), r)
    assert_allclose(m_shape.pixels, m_mask.pixels)


def test_rescale_boolean():
    mask = BooleanImage.blank((100, 100))
    mask.resize((10, 10))



from __future__ import division
from functools import wraps
import numpy as np
from menpo.image import Image, MaskedImage, BooleanImage
from menpo.transform import Translation, NonUniformScale


def lm_centres_correction(centres):
    r"""
    Construct a transform that will correct landmarks for a window
    iterating feature calculation

    Parameters
    ----------
    centres : `ndarray` (H, W, 2)
        The location of the window centres in the features

    Returns
    -------
    :map:`Affine`
        An affine transform that performs the correction.
        Should be applied to the landmarks on the target image.
    """
    t = Translation(-centres.min(axis=0).min(axis=0), skip_checks=True)
    step_v = centres[0, 0, 0]
    if centres.shape[0] > 1:
        step_v = centres[1, 0, 0] - centres[0, 0, 0]
    step_h = centres[0, 0, 1]
    if centres.shape[1] > 1:
        step_h = centres[0, 1, 1] - centres[0, 0, 1]
    s = NonUniformScale((1./step_v, 1./step_h), skip_checks=True)
    return t.compose_before(s)


def sample_mask_for_centres(mask, centres):
    r"""
    Sample a mask at the centres

    Parameters
    ----------
    mask :  Either MaskedImage or Image class.
        The target image object that includes the windows_centres.

    window_centres : ndarray, optional
        If set, use these window centres to rescale the landmarks
        appropriately. If None, no scaling is applied.

    """
    return BooleanImage(mask[centres[..., 0], centres[..., 1]], copy=False)


def rebuild_feature_image(image, f_pixels):
    shape_changed = f_pixels.shape[:-1] != image.shape
    if hasattr(image, 'mask'):
        # original image had a mask. Did the feature generate an image of the
        # same size?
        if shape_changed:
            # feature is of a different size - best we can do is rescale the
            # mask
            mask = image.mask.resize(f_pixels.shape[:-1])
        else:
            # feature is same size as input
            mask = image.mask.copy()
        new_image = MaskedImage(f_pixels, mask=mask, copy=False)
    else:
        new_image = Image(f_pixels, copy=False)
    if image.has_landmarks:
        if shape_changed:
            # need to adjust the landmarks
            sf = np.array(f_pixels.shape[:-1]) / np.array(image.shape)
            new_image.landmarks = NonUniformScale(sf).apply(image.landmarks)
        else:
            new_image.landmarks = image.landmarks
    return new_image


def rebuild_feature_image_with_centres(image, f_pixels, centres):
    if hasattr(image, 'mask'):
        mask = sample_mask_for_centres(image.mask.mask, centres)
        new_image = MaskedImage(f_pixels, mask=mask, copy=False)
    else:
        new_image = Image(f_pixels, copy=False)
    if image.has_landmarks:
        t = lm_centres_correction(centres)
        new_image.landmarks = t.apply(image.landmarks)
    return new_image


def imgfeature(wrapped):

    @wraps(wrapped)
    def wrapper(image, *args, **kwargs):
        if isinstance(image, np.ndarray):
            # ndarray supplied to Image feature - build a
            # temp image for it and just return the pixels
            image = Image(image, copy=False)
            return wrapped(image, *args, **kwargs).pixels
        else:
            return wrapped(image, *args, **kwargs)
    return wrapper


def ndfeature(wrapped):

    @wraps(wrapped)
    def wrapper(image, *args, **kwargs):
        if not isinstance(image, np.ndarray):
            # Image supplied to ndarray feature -
            # extract pixels and go
            feature = wrapped(image.pixels, *args, **kwargs)
            return rebuild_feature_image(image, feature)
        else:
            return wrapped(image, *args, **kwargs)
    return wrapper


def winitfeature(wrapped):

    @wraps(wrapped)
    def wrapper(image, *args, **kwargs):
        if not isinstance(image, np.ndarray):
            # Image supplied to ndarray feature -
            # extract pixels and go
            feature, centres = wrapped(image.pixels, *args, **kwargs)
            return rebuild_feature_image_with_centres(image, feature, centres)
        else:
            # user just supplied ndarray - give them ndarray back
            return wrapped(image, *args, **kwargs)[0]

    return wrapper



import numpy as np
from numpy.testing import assert_allclose

from menpo.base import Vectorizable
from menpo.transform.base import VInvertible


ones_vector = np.ones(3)


class MockedVInvertable(VInvertible, Vectorizable):
    def __init__(self):
        self.vector = ones_vector

    def from_vector_inplace(self, vector):
        self.vector = vector

    def _as_vector(self):
        return self.vector

    def pseudoinverse(self):
        m = MockedVInvertable()
        m.vector = -self.vector
        return m

    def has_true_inverse(self):
        return True


def vinertable_pseudoinverse_test():
    v = MockedVInvertable()
    inv = v.pseudoinverse()
    assert_allclose(inv.vector, -ones_vector)


def vinertable_pseudoinverse_vector_test():
    v = MockedVInvertable()
    arr = np.array([1, 2])
    vec = v.pseudoinverse_vector(arr)
    assert_allclose(vec, -arr)



from menpo.model.instancebacked import InstanceBackedModel
from menpo.model.linear import LinearModel, MeanLinearModel


class InstanceLinearModel(LinearModel, InstanceBackedModel):
    r"""
    Mixin of :map:`LinearModel` and :map:`InstanceBackedModel` objects.

    Parameters
    ----------
    components : ``(n_components, n_features)`` `ndarray`
        The components array.
    template_instance : :map:`Vectorizable`
        The template instance.
    """

    def __int__(self, components, template_instance):
        LinearModel.__init__(self, components)
        InstanceBackedModel.__init__(self, template_instance)


class MeanInstanceLinearModel(MeanLinearModel, InstanceBackedModel):
    r"""
    Mixin of :map:`MeanLinearModel` and :map:`InstanceBackedModel` objects.

    Parameters
    ----------
    components : ``(n_components, n_features)`` `ndarray`
        The components array.
    mean_vector : ``(n_features,)`` `ndarray`
        The mean vector.
    template_instance : :map:`Vectorizable`
        The template instance.
    """

    def __init__(self, components, mean_vector, template_instance):
        MeanLinearModel.__init__(self, components, mean_vector)
        InstanceBackedModel.__init__(self, template_instance)

    def mean(self):
        r"""
        Return the mean of the model.

        :type: :map:`Vectorizable`
        """
        return self.template_instance.from_vector(self.mean_vector)

    def component(self, index, with_mean=True, scale=1.0):
        r"""
        Return a particular component of the linear model.

        Parameters
        ----------
        index : `int`
            The component that is to be returned
        with_mean: `bool`, optional
            If ``True``, the component will be blended with the mean vector
            before being returned. If not, the component is returned on it's
            own.
        scale : `float`, optional
            A scale factor that should be applied to the component. Only
            valid in the case where ``with_mean == True``. See
            :meth:`component_vector` for how this scale factor is interpreted.

        Returns
        -------
        component : `type(self.template_instance)`
            The requested component.
        """
        return self.template_instance.from_vector(self.component_vector(
            index, with_mean=with_mean, scale=scale))



from .base import (
    Renderer, Viewable, LandmarkableViewable, viewwrapper, Menpo3dErrorMessage,
    PointGraphViewer2d, LandmarkViewer2d, ImageViewer2d, ImageViewer,
    AlignmentViewer2d, GraphPlotter, view_image_landmarks)
from .text_utils import progress_bar_str, print_dynamic, print_bytes
from .widgets import (visualize_pointclouds, visualize_images,
                      visualize_landmarks, features_selection,
                      save_matplotlib_figure, visualize_landmarkgroups)
from .viewmatplotlib import MatplotlibRenderer



import os
import sys
from setuptools import setup, find_packages
import versioneer


on_rtd = os.environ.get('READTHEDOCS', None) == 'True'

if on_rtd:
    install_requires = []
    ext_modules = []
    include_dirs = []
    cython_exts = []
else:
    from Cython.Build import cythonize
    import numpy as np

    # ---- C/C++ EXTENSIONS ---- #
    cython_modules = ['menpo/shape/mesh/normals.pyx',
                      'menpo/transform/piecewiseaffine/fastpwa.pyx',
                      'menpo/feature/windowiterator.pyx',
                      'menpo/external/skimage/_warps_cy.pyx',
                      'menpo/image/extract_patches.pyx']

    cython_exts = cythonize(cython_modules, quiet=True)
    include_dirs = [np.get_include()]
    install_requires = ['numpy>=1.9.1,<1.10',
                        'scipy>=0.15,<0.16',
                        'matplotlib>=1.4,<1.5',
                        'pillow==2.7.0',
                        'Cython>=0.21,<0.22']

    if sys.version_info.major == 2:
        install_requires.append('pathlib==1.0')

# Versioneer allows us to automatically generate versioning from
# our git tagging system which makes releases simpler.
versioneer.VCS = 'git'
versioneer.versionfile_source = 'menpo/_version.py'
versioneer.versionfile_build = 'menpo/_version.py'
versioneer.tag_prefix = 'v'  # tags are like v1.2.0
versioneer.parentdir_prefix = 'menpo-'  # dirname like 'menpo-v1.2.0'

setup(name='menpo',
      version=versioneer.get_version(),
      cmdclass=versioneer.get_cmdclass(),
      description='iBUG Facial Modelling Toolkit',
      author='James Booth',
      author_email='james.booth08@imperial.ac.uk',
      include_dirs=include_dirs,
      ext_modules=cython_exts,
      packages=find_packages(),
      install_requires=install_requires,
      package_data={'menpo': ['data/*',
                              'feature/cpp/*.cpp',
                              'feature/cpp/*.h',
                              'transform/piecewiseaffine/fastpwa/*.c',
                              'transform/piecewiseaffine/fastpwa/*.h'],
                    '': ['*.pxd', '*.pyx']},
      tests_require=['nose', 'mock']
)



from .features import (gradient, hog, lbp, es, igo, no_op, gaussian_filter,
                       daisy, features_selection_widget)
from .predefined import sparse_hog, double_igo
from .base import ndfeature, imgfeature



import numpy as np
import PIL.Image as PILImage
from .base import Importer
from menpo.image import Image, MaskedImage, BooleanImage


class PILImporter(Importer):
    r"""
    Imports an image using PIL.

    Different image modes cause different importing strategies.

    RGB, L, I:
        Imported as either `float` or `uint8` depending on normalisation flag.
    RGBA:
        Imported as :map:`MaskedImage` if normalise is ``True`` else imported
        as a 4 channel `uint8` image.
    1:
        Imported as a :map:`BooleanImage`. Normalisation is ignored.
    F:
        Imported as a floating point image. Normalisation is ignored.

    Parameters
    ----------
    filepath : string
        Absolute filepath of image
    normalise : `bool`, optional
        If ``True``, normalise between 0.0 and 1.0 and convert to float. If
        ``False`` just pass whatever PIL imports back (according
        to types rules outlined in constructor).
    """
    def __init__(self, filepath, normalise=True):
        super(PILImporter, self).__init__(filepath)
        self._pil_image = None
        self.normalise = normalise

    def build(self):
        r"""
        Read the image using PIL and then use the :map:`Image` constructor to
        create a class.
        """
        self._pil_image = PILImage.open(self.filepath)
        mode = self._pil_image.mode
        if mode == 'RGBA':
            # RGB with Alpha Channel
            # If we normalise it then we convert to floating point
            # and set the alpha channel to the mask
            if self.normalise:
                alpha = np.array(self._pil_image)[..., 3].astype(np.bool)
                image_pixels = self._pil_to_numpy(True,
                                                  convert='RGB')
                image = MaskedImage(image_pixels, mask=alpha)
            else:
                # With no normalisation we just return the pixels
                image = Image(self._pil_to_numpy(False))
        elif mode in ['L', 'I', 'RGB']:
            # Greyscale, Integer and RGB images
            image = Image(self._pil_to_numpy(self.normalise))
        elif mode == '1':
            # Can't normalise a binary image
            image = BooleanImage(self._pil_to_numpy(False))
        elif mode == 'P':
            # Convert pallete images to RGB
            image = Image(self._pil_to_numpy(self.normalise, convert='RGB'))
        elif mode == 'F':  # Floating point images
            # Don't normalise as we don't know the scale
            image = Image(self._pil_to_numpy(False))
        else:
            raise ValueError('Unexpected mode for PIL: {}'.format(mode))
        return image

    def _pil_to_numpy(self, normalise, convert=None):
        dtype = np.float if normalise else None
        p = self._pil_image.convert(convert) if convert else self._pil_image
        np_pixels = np.array(p, dtype=dtype, copy=True)
        return np_pixels / 255.0 if normalise else np_pixels


class PILGIFImporter(PILImporter):
    r"""
    Imports a GIF using PIL. Correctly encodes the pallete
    (`P` for `RGB` (Pallete mode).

    For multi-frame GIF animations, will return a list of images containing
    each frame.

    Parameters
    ----------
    filepath : string
        Absolute filepath of image
    normalise : `bool`, optional
        If ``True``, normalise between 0.0 and 1.0 and convert to float. If
        ``False`` just pass whatever PIL imports back (according
        to types rules outlined in constructor).
    """

    def __init__(self, filepath, normalise=True):
        super(PILGIFImporter, self).__init__(filepath, normalise=normalise)

    def build(self):
        r"""
        Read the image using PIL and then use the :map:`Image` constructor to
        create a class.
        """
        self._pil_image = PILImage.open(self.filepath)
        # By default GIFs use a
        if self._pil_image.mode == 'P':
            # Do we need this duration information for playback?
            # duration = self._pil_image.info['duration']
            images = []
            try:
                while 1:  # Keep looping until we hit the end of the GIF
                    np_pixels = self._pil_to_numpy(self.normalise,
                                                   convert='RGB')
                    images.append(Image(np_pixels))
                    # Seek to the next frame
                    self._pil_image.seek(self._pil_image.tell() + 1)
            except EOFError:
                pass  # Exhausted GIF
        else:
            raise ValueError('Unknown mode for GIF: {}'.format(
                self._pil_image.mode))

        return images



import numpy as np

from ..adjacency import mask_adjacency_array, reindex_adjacency_array
from .base import TriMesh


class ColouredTriMesh(TriMesh):
    r"""
    Combines a :map:`TriMesh` with a colour per vertex.

    Parameters
    ----------
    points : ``(n_points, n_dims)`` `ndarray`
        The array representing the points.
    trilist : ``(M, 3)`` `ndarray` or ``None``, optional
        The triangle list. If `None`, a Delaunay triangulation of
        the points will be used instead.
    colours : ``(N, 3)`` `ndarray`, optional
        The floating point RGB colour per vertex. If not given, grey will be
        assigned to each vertex.
    copy: `bool`, optional
        If ``False``, the points, trilist and colours will not be copied on
        assignment.
        In general this should only be used if you know what you are doing.

    Raises
    ------
    ValueError
        If the number of colour values does not match the number of vertices.
    """

    def __init__(self, points, trilist=None, colours=None, copy=True):
        TriMesh.__init__(self, points, trilist=trilist, copy=copy)
        # Handle the settings of colours, either be provided a default grey
        # set of colours, or copy the given array if necessary
        if colours is None:
            # default to grey
            colours_handle = np.ones_like(points, dtype=np.float) * 0.5
        elif not copy:
            colours_handle = colours
        else:
            colours_handle = colours.copy()

        if points.shape[0] != colours.shape[0]:
            raise ValueError('Must provide a colour per-vertex.')
        self.colours = colours_handle

    def from_mask(self, mask):
        """
        A 1D boolean array with the same number of elements as the number of
        points in the ColouredTriMesh. This is then broadcast across the
        dimensions of the mesh and returns a new mesh containing only those
        points that were ``True`` in the mask.

        Parameters
        ----------
        mask : ``(n_points,)`` `ndarray`
            1D array of booleans

        Returns
        -------
        mesh : :map:`ColouredTriMesh`
            A new mesh that has been masked.
        """
        if mask.shape[0] != self.n_points:
            raise ValueError('Mask must be a 1D boolean array of the same '
                             'number of entries as points in this '
                             'ColouredTriMesh.')

        ctm = self.copy()
        if np.all(mask):  # Fast path for all true
            return ctm
        else:
            # Recalculate the mask to remove isolated vertices
            isolated_mask = self._isolated_mask(mask)
            # Recreate the adjacency array with the updated mask
            masked_adj = mask_adjacency_array(isolated_mask, self.trilist)
            ctm.trilist = reindex_adjacency_array(masked_adj)
            ctm.points = ctm.points[isolated_mask, :]
            ctm.colours = ctm.colours[isolated_mask, :]
            return ctm

    def _view_3d(self, figure_id=None, new_figure=False, coloured=True,
                 **kwargs):
        r"""
        Visualize the :map:`ColouredTriMesh` in 3D.

        Parameters
        ----------
        figure_id : `object`, optional
            The id of the figure to be used.
        new_figure : `bool`, optional
            If ``True``, a new figure is created.
        coloured : `bool`, optional
            If `True`, render the colours.

        Returns
        -------
        viewer : :map:`Renderer`
            The viewer object.
        """
        if coloured:
            try:
                from menpo3d.visualize import ColouredTriMeshViewer3d
                return ColouredTriMeshViewer3d(
                    figure_id, new_figure, self.points,
                    self.trilist, self.colours).render(**kwargs)
            except ImportError:
                from menpo.visualize import Menpo3dErrorMessage
                raise ImportError(Menpo3dErrorMessage)
        else:
            return super(ColouredTriMesh, self).view(figure_id=figure_id,
                                                     new_figure=new_figure,
                                                     **kwargs)

    def _view_2d(self, figure_id=None, new_figure=False, image_view=True,
                 render_lines=True, line_colour='r', line_style='-',
                 line_width=1., render_markers=True, marker_style='o',
                 marker_size=20, marker_face_colour='k', marker_edge_colour='k',
                 marker_edge_width=1., render_axes=True,
                 axes_font_name='sans-serif', axes_font_size=10,
                 axes_font_style='normal', axes_font_weight='normal',
                 axes_x_limits=None, axes_y_limits=None, figure_size=(10, 8),
                 label=None):
        r"""
        Visualization of the TriMesh in 2D. Currently, explicit coloured TriMesh
        viewing is not supported, and therefore viewing falls back to uncoloured
        2D TriMesh viewing.

        Returns
        -------
        figure_id : `object`, optional
            The id of the figure to be used.
        new_figure : `bool`, optional
            If ``True``, a new figure is created.
        image_view : `bool`, optional
            If ``True`` the ColouredTriMesh will be viewed as if it is in the
            image coordinate system.
        render_lines : `bool`, optional
            If ``True``, the edges will be rendered.
        line_colour : See Below, optional
            The colour of the lines.
            Example options::

                {r, g, b, c, m, k, w}
                or
                (3, ) ndarray

        line_style : ``{-, --, -., :}``, optional
            The style of the lines.
        line_width : `float`, optional
            The width of the lines.
        render_markers : `bool`, optional
            If ``True``, the markers will be rendered.
        marker_style : See Below, optional
            The style of the markers. Example options ::

                {., ,, o, v, ^, <, >, +, x, D, d, s, p, *, h, H, 1, 2, 3, 4, 8}

        marker_size : `int`, optional
            The size of the markers in points^2.
        marker_face_colour : See Below, optional
            The face (filling) colour of the markers.
            Example options ::

                {r, g, b, c, m, k, w}
                or
                (3, ) ndarray

        marker_edge_colour : See Below, optional
            The edge colour of the markers.
            Example options ::

                {r, g, b, c, m, k, w}
                or
                (3, ) ndarray

        marker_edge_width : `float`, optional
            The width of the markers' edge.
        render_axes : `bool`, optional
            If ``True``, the axes will be rendered.
        axes_font_name : See Below, optional
            The font of the axes.
            Example options ::

                {serif, sans-serif, cursive, fantasy, monospace}

        axes_font_size : `int`, optional
            The font size of the axes.
        axes_font_style : {``normal``, ``italic``, ``oblique``}, optional
            The font style of the axes.
        axes_font_weight : See Below, optional
            The font weight of the axes.
            Example options ::

                {ultralight, light, normal, regular, book, medium, roman,
                semibold, demibold, demi, bold, heavy, extra bold, black}

        axes_x_limits : (`float`, `float`) `tuple` or ``None``, optional
            The limits of the x axis.
        axes_y_limits : (`float`, `float`) `tuple` or ``None``, optional
            The limits of the y axis.
        figure_size : (`float`, `float`) `tuple` or ``None``, optional
            The size of the figure in inches.
        label : `str`, optional
            The name entry in case of a legend.

        Returns
        -------
        viewer : :map:`PointGraphViewer2d`
            The viewer object.

        Raises
        ------
        warning
            2D Viewing of Coloured TriMeshes is not supported, automatically
            falls back to 2D :map:`TriMesh` viewing.
        """
        import warnings
        warnings.warn(Warning('2D Viewing of Coloured TriMeshes is not '
                              'supported, falling back to TriMesh viewing.'))
        return TriMesh._view_2d(
            self, figure_id=figure_id, new_figure=new_figure,
            image_view=image_view, render_lines=render_lines,
            line_colour=line_colour, line_style=line_style,
            line_width=line_width, render_markers=render_markers,
            marker_style=marker_style, marker_size=marker_size,
            marker_face_colour=marker_face_colour,
            marker_edge_colour=marker_edge_colour,
            marker_edge_width=marker_edge_width, render_axes=render_axes,
            axes_font_name=axes_font_name, axes_font_size=axes_font_size,
            axes_font_style=axes_font_style, axes_font_weight=axes_font_weight,
            axes_x_limits=axes_x_limits, axes_y_limits=axes_y_limits,
            figure_size=figure_size, label=label)
